# The MIT License (MIT)
# Copyright (c) 2014 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Classes and enums for documents in the Azure Cosmos database service.
"""

from typing import Optional, TYPE_CHECKING, Union, Any

from typing_extensions import Literal, TypedDict

from ._retry_options import RetryOptions

if TYPE_CHECKING:
    from ._retry_utility import ConnectionRetryPolicy


class UserConsistencyPolicy(TypedDict, total=False):
    defaultConsistencyLevel: str
    """The default consistency level."""
    maxStalenessPrefix: int
    """In bounded staleness consistency, the maximum allowed staleness in
    terms difference in sequence numbers (aka version)."""
    maxStalenessIntervalInSeconds: int
    """In bounded staleness consistency, the maximum allowed staleness in
    terms time interval."""


class DatabaseAccount:  # pylint: disable=too-many-instance-attributes
    """Database account.

    A DatabaseAccount is the container for databases.

    :ivar str DatabaseLink:
        The self-link for Databases in the databaseAccount.
    :ivar str MediaLink:
        The self-link for Media in the databaseAccount.
    :ivar int MaxMediaStorageUsageInMB:
        Attachment content (media) storage quota in MBs (Retrieved from gateway).
    :ivar int CurrentMediaStorageUsageInMB:
        Current attachment content (media) usage in MBs (Retrieved from gateway).
        Value is returned from cached information updated periodically and
        is not guaranteed to be real time.
    :ivar ConsistencyPolicy:
        UserConsistencyPolicy settings.
    :vartype ConsistencyPolicy:
        dict[str, Any]
    :ivar boolean EnableMultipleWritableLocations:
        Flag on the azure Cosmos account that indicates if writes can take
        place in multiple locations.
    """

    def __init__(self) -> None:
        self.DatabasesLink: str = ""
        self.MediaLink: str = ""
        self.MaxMediaStorageUsageInMB: int = 0
        self.CurrentMediaStorageUsageInMB: int = 0
        self.ConsumedDocumentStorageInMB: int = 0
        self.ReservedDocumentStorageInMB: int = 0
        self.ProvisionedDocumentStorageInMB: int = 0
        self.ConsistencyPolicy: Optional[dict[str, Any]] = None
        self._WritableLocations: list[dict[str, str]] = []
        self._ReadableLocations: list[dict[str, str]] = []
        self._EnableMultipleWritableLocations = False

    @property
    def WritableLocations(self) -> list[dict[str, str]]:
        """The list of writable locations for a geo-replicated database account.
        :returns: List of writable locations for the database account.
        :rtype: list[dict[str, str]]
        """
        return self._WritableLocations

    @property
    def ReadableLocations(self) -> list[dict[str, str]]:
        """The list of readable locations for a geo-replicated database account.
        :returns: List of readable locations for the database account.
        :rtype: list[dict[str, str]]
        """
        return self._ReadableLocations


class ConsistencyLevel:
    """Represents the consistency levels supported for Azure Cosmos client
    operations.

    The requested ConsistencyLevel must match or be weaker than that provisioned
    for the database account. Consistency levels.

    Consistency levels by order of strength are Strong, BoundedStaleness,
    Session, ConsistentPrefix and Eventual.
    """
    Strong: Literal["Strong"] = "Strong"
    """Strong Consistency guarantees that read operations always return the
    value that was last written.
    """
    BoundedStaleness: Literal["BoundedStaleness"] = "BoundedStaleness"
    """Bounded Staleness guarantees that reads are not too out-of-date. This
    can be configured based on number of operations (MaxStalenessPrefix)
    or time (MaxStalenessIntervalInSeconds).
    """
    Session: Literal["Session"] = "Session"
    """Session Consistency guarantees monotonic reads (you never read old data,
    then new, then old again), monotonic writes (writes are ordered) and
    read your writes (your writes are immediately visible to your reads)
    within any single session.
    """
    Eventual: Literal["Eventual"] = "Eventual"
    """Eventual Consistency guarantees that reads will return a subset of
    writes. All writes will be eventually be available for reads.
    """
    ConsistentPrefix: Literal["ConsistentPrefix"] = "ConsistentPrefix"
    """ConsistentPrefix Consistency guarantees that reads will return some
    prefix of all writes with no gaps. All writes will be eventually be
    available for reads.
    """


class IndexingMode:
    """Specifies the supported indexing modes."""
    Consistent: Literal["consistent"] = "consistent"
    """Index is updated synchronously with a create or update operation. With
    consistent indexing, query behavior is the same as the default
    consistency level for the collection.
    The index is always kept up to date with the data.
    """
    Lazy: Literal["lazy"] = "lazy"
    """Index is updated asynchronously with respect to a create or update
    operation. Not supported for new containers since June/2020.
    With lazy indexing, queries are eventually consistent. The index is
    updated when the collection is idle.
    """
    NoIndex: Literal["none"] = "none"
    """No index is provided.
    Setting IndexingMode to "None" drops the index. Use this if you don't
    want to maintain the index for a document collection, to save the
    storage cost or improve the write throughput. Your queries will
    degenerate to scans of the entire collection.
    """


class IndexKind:
    """Specifies the index kind of index specs."""
    Hash: Literal["Hash"] = "Hash"
    """The index entries are hashed to serve point look up queries.
    Can be used to serve queries like: SELECT * FROM docs d WHERE d.prop = 5
    """
    Range: Literal["Range"] = "Range"
    """
    The index entries are ordered. Range indexes are optimized for
    inequality predicate queries with efficient range scans.
    Can be used to serve queries like: SELECT * FROM docs d WHERE d.prop > 5
    """
    MultiHash: Literal["MultiHash"] = "MultiHash"
    """MultiHash"""


class PartitionKind:
    """Specifies the kind of partitioning to be applied."""
    Hash: Literal["Hash"] = "Hash"
    """The partition key definition path is hashed."""
    MultiHash: Literal["MultiHash"] = "MultiHash"
    """MultiHash"""


class DataType:
    """Specifies the data type of index specs."""
    Number: Literal["Number"] = "Number"
    """Represents a numeric data type."""
    String: Literal["String"] = "String"
    """Represents a string data type."""
    Point: Literal["Point"] = "Point"
    """Represents a point data type."""
    LineString: Literal["LineString"] = "LineString"
    """Represents a line string data type."""
    Polygon: Literal["Polygon"] = "Polygon"
    """Represents a polygon data type."""
    MultiPolygon: Literal["MultiPolygon"] = "MultiPolygon"
    """Represents a multi-polygon data type."""


class IndexingDirective:
    """Specifies whether the resource is to be indexed."""
    Default: int = 0
    """Use any pre-defined/pre-configured defaults."""
    Exclude: int = 1
    """Index the resource."""
    Include: int = 2
    """Do not index the resource."""


class ConnectionMode:
    """Represents the connection mode to be used by the client."""
    Gateway: int = 0
    """Use the Azure Cosmos gateway to route all requests. The gateway proxies
    requests to the right data partition.
    """


class PermissionMode:
    """Applicability of a permission."""
    NoneMode: Literal["none"] = "none"  # None is python's key word.
    """None"""
    Read: Literal["read"] = "read"
    """Permission applicable for read operations only."""
    All: Literal["all"] = "all"
    """Permission applicable for all operations."""


class TriggerType:
    """Specifies the type of trigger."""
    Pre: Literal["pre"] = "pre"
    """Trigger should be executed before the associated operation(s)."""
    Post: Literal["post"] = "post"
    """Trigger should be executed after the associated operation(s)."""


class TriggerOperation:
    """Specifies the operations on which a trigger should be executed."""
    All: Literal["all"] = "all"
    """All operations."""
    Create: Literal["create"] = "create"
    """Create operations only."""
    Update: Literal["update"] = "update"
    """Update operations only."""
    Delete: Literal["delete"] = "delete"
    """Delete operations only."""
    Replace: Literal["replace"] = "replace"
    """Replace operations only."""


class SSLConfiguration:
    """Configuration for SSL connections.

    See https://requests.readthedocs.io/en/master/user/advanced/#ssl-cert-verification
    for more information.

    :ivar str SSLKeyFIle:
        The path of the key file for ssl connection.
    :ivar str SSLCertFile:
        The path of the cert file for ssl connection.
    :ivar SSLCaCerts:
        The path of the CA_BUNDLE file with certificates of trusted CAs.
    :vartype SSLCaCerts: str or bool
    """
    def __init__(self) -> None:
        self.SSLKeyFile: Optional[str] = None
        self.SSLCertFile: Optional[str] = None
        self.SSLCaCerts: Optional[Union[str, bool]] = None


class ProxyConfiguration:
    """Configuration for a proxy.

    :ivar str Host:
        The host address of the proxy.
    :ivar int Port:
        The port number of the proxy.
    """
    def __init__(self) -> None:
        self.Host: Optional[str] = None
        self.Port: Optional[int] = None


class ConnectionPolicy:  # pylint: disable=too-many-instance-attributes
    """Represents the Connection policy associated with a CosmosClientConnection.

    :ivar int RequestTimeout:
        Gets or sets the request timeout (time to wait for a response from a
        network peer).
    :ivar ConnectionMode:
        Gets or sets the connection mode used in the client. (Currently only
        Gateway is supported.)
    :vartype ConnectionMode: ~azure.cosmos.documents.ConnectionMode
    :ivar SSLConfiguration:
        Gets or sets the SSL configuration.
    :vartype SSLConfiguration: ~azure.cosmos.documents.SSLConfiguration
    :ivar ProxyConfiguration:
        Gets or sets the proxy configuration.
    :vartype ProxyConfiguration: ~azure.cosmos.documents.ProxyConfiguration
    :ivar boolean EnableEndpointDiscovery:
        Gets or sets endpoint discovery flag for geo-replicated database
        accounts. When EnableEndpointDiscovery is true, the client will
        automatically discover the current write and read locations and direct
        the requests to the correct location taking into consideration of the
        user's preference(if provided) as PreferredLocations.
    :ivar PreferredLocations:
        Gets or sets the preferred locations for geo-replicated database
        accounts. When EnableEndpointDiscovery is true and PreferredLocations is
        non-empty, the client will use this list to evaluate the final location,
        taking into consideration the order specified in PreferredLocations. The
        locations in this list are specified as the names of the azure Cosmos
        locations like, 'West US', 'East US', 'Central India' and so on.
    :vartype PreferredLocations: list[str]
    :ivar ExcludedLocations:
        Gets or sets the excluded locations for geo-replicated database
        accounts. When ExcludedLocations is non-empty, the client will skip this
        set of locations from the final location evaluation. The locations in
        this list are specified as the names of the azure Cosmos locations like,
        'West US', 'East US', 'Central India' and so on.
    :vartype ExcludedLocations: list[str]
    :ivar RetryOptions:
        Gets or sets the retry options to be applied to all requests when
        retrying.
    :vartype RetryOptions: ~RetryOptions
    :ivar boolean DisableSSLVerification:
        Flag to disable SSL verification for the requests. SSL verification is
        enabled by default.
        This is intended to be used only when targeting emulator endpoint to
        avoid failing your requests with SSL related error.
        DO NOT set this when targeting production endpoints.
    :ivar boolean UseMultipleWriteLocations:
        Flag to enable writes on any locations (regions) for geo-replicated
        database accounts in the Azure Cosmos database service.
    :ivar ConnectionRetryConfiguration:
        Retry Configuration to be used for connection retries.
    :vartype ConnectionRetryConfiguration:
        int or ~azure.cosmos.ConnectionRetryPolicy
    :ivar boolean ResponsePayloadOnWriteDisabled:
        Indicates whether service should be instructed to skip sending response payloads.
    :ivar int RetryNonIdempotentWrites:
        Indicates how many times the client should retry non-idempotent write requests for item operations.
    """

    __defaultRequestTimeout: int = 5  # seconds
    __defaultDBAConnectionTimeout: int = 3  # seconds
    __defaultReadTimeout: int = 65  # seconds
    __defaultRecoveryReadTimeout: int = 6  # seconds
    __defaultDBAReadTimeout: int = 3 # seconds
    __defaultMaxBackoff: int = 1 # seconds

    def __init__(self) -> None:
        # RequestTimeout is the connection timeout for all operations except database account
        self.RequestTimeout: int = self.__defaultRequestTimeout
        self.DBAConnectionTimeout: int = self.__defaultDBAConnectionTimeout
        self.ReadTimeout: int = self.__defaultReadTimeout
        # The request timeout for a request trying to recover a unavailable partition
        # This is only applicable if circuit breaker is enabled
        self.RecoveryReadTimeout: int = self.__defaultRecoveryReadTimeout
        self.DBAReadTimeout: int = self.__defaultDBAReadTimeout
        self.MaxBackoff: int = self.__defaultMaxBackoff
        self.ConnectionMode: int = ConnectionMode.Gateway
        self.SSLConfiguration: Optional[SSLConfiguration] = None
        self.ProxyConfiguration: Optional[ProxyConfiguration] = None
        self.EnableEndpointDiscovery: bool = True
        self.PreferredLocations: list[str] = []
        self.ExcludedLocations: Optional[list[str]] = None
        self.RetryOptions: RetryOptions = RetryOptions()
        self.DisableSSLVerification: bool = False
        self.UseMultipleWriteLocations: bool = False
        self.ConnectionRetryConfiguration: Optional["ConnectionRetryPolicy"] = None
        self.ResponsePayloadOnWriteDisabled: bool = False
        self.RetryNonIdempotentWrites: int = 0

class _OperationType:
    """Represents the type of the operation"""
    Create: Literal["Create"] = "Create"
    Delete: Literal["Delete"] = "Delete"
    ExecuteJavaScript: Literal["ExecuteJavaScript"] = "ExecuteJavaScript"
    Head: Literal["Head"] = "Head"
    HeadFeed: Literal["HeadFeed"] = "HeadFeed"
    Patch: Literal["Patch"] = "Patch"
    Query: Literal["Query"] = "Query"
    QueryPlan: Literal["QueryPlan"] = "QueryPlan"
    Read: Literal["Read"] = "Read"
    ReadFeed: Literal["ReadFeed"] = "ReadFeed"
    Recreate: Literal["Recreate"] = "Recreate"
    Replace: Literal["Replace"] = "Replace"
    SqlQuery: Literal["SqlQuery"] = "SqlQuery"
    Update: Literal["Update"] = "Update"
    Upsert: Literal["Upsert"] = "Upsert"
    Batch: Literal["Batch"] = "Batch"

    @staticmethod
    def IsWriteOperation(operationType: str) -> bool:
        return operationType in (
            _OperationType.Create,
            _OperationType.Delete,
            _OperationType.Recreate,
            _OperationType.ExecuteJavaScript,
            _OperationType.Replace,
            _OperationType.Upsert,
            _OperationType.Update,
            _OperationType.Batch,
            _OperationType.Patch
        )

    @staticmethod
    def IsReadOnlyOperation(operationType: str) -> bool:
        return operationType in (
            _OperationType.Read,
            _OperationType.ReadFeed,
            _OperationType.Head,
            _OperationType.HeadFeed,
            _OperationType.Query,
            _OperationType.SqlQuery,
            _OperationType.QueryPlan
        )

class _QueryFeature:
    NoneQuery: Literal["NoneQuery"] = "NoneQuery"
    Aggregate: Literal["Aggregate"] = "Aggregate"
    CompositeAggregate: Literal["CompositeAggregate"] = "CompositeAggregate"
    Distinct: Literal["Distinct"] = "Distinct"
    GroupBy: Literal["GroupBy"] = "GroupBy"
    MultipleAggregates: Literal["MultipleAggregates"] = "MultipleAggregates"
    MultipleOrderBy: Literal["MultipleOrderBy"] = "MultipleOrderBy"
    OffsetAndLimit: Literal["OffsetAndLimit"] = "OffsetAndLimit"
    OrderBy: Literal["OrderBy"] = "OrderBy"
    Top: Literal["Top"] = "Top"
    NonStreamingOrderBy: Literal["NonStreamingOrderBy"] = "NonStreamingOrderBy"
    HybridSearch: Literal["HybridSearch"] = "HybridSearch"
    CountIf: Literal["CountIf"] = "CountIf"
    WeightedRankFusion: Literal["WeightedRankFusion"] = "WeightedRankFusion"

class _DistinctType:
    NoneType: Literal["None"] = "None"
    Ordered: Literal["Ordered"] = "Ordered"
    Unordered: Literal["Unordered"] = "Unordered"



# The MIT License (MIT)
# Copyright (c) 2021 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Create, read, and delete databases in the Azure Cosmos DB SQL API service.
"""

from typing import Any, Optional, Union, cast, Mapping, Iterable, Callable, overload, Literal
import warnings
from azure.core.async_paging import AsyncItemPaged
from azure.core.credentials import TokenCredential
from azure.core.credentials_async import AsyncTokenCredential
from azure.core.pipeline.policies import RetryMode

from azure.core.tracing.decorator_async import distributed_trace_async
from azure.core.tracing.decorator import distributed_trace
from azure.cosmos.offer import ThroughputProperties

from ..cosmos_client import _parse_connection_str
from .._constants import _Constants as Constants
from ._cosmos_client_connection_async import CosmosClientConnection, CredentialDict
from .._base import build_options as _build_options, _set_throughput_options
from ._retry_utility_async import _ConnectionRetryPolicy
from ._database import DatabaseProxy, _get_database_link
from ..documents import ConnectionPolicy, DatabaseAccount
from ..exceptions import CosmosResourceNotFoundError
from .._cosmos_responses import CosmosDict

# pylint: disable=docstring-keyword-should-match-keyword-only

__all__ = ("CosmosClient",)

CredentialType = Union[
    AsyncTokenCredential, CredentialDict, str, Mapping[str, Any], Iterable[Mapping[str, Any]]
]

def _build_auth(credential: CredentialType) -> CredentialDict:
    auth: CredentialDict = {}
    if isinstance(credential, str):
        auth['masterKey'] = credential
    elif isinstance(credential, Mapping):
        if any(k for k in credential.keys() if k in ['masterKey', 'resourceTokens', 'permissionFeed']):
            return cast(CredentialDict, credential)  # Backwards compatible
        auth['resourceTokens'] = credential
    elif isinstance(credential, Iterable):
        auth['permissionFeed'] = cast(Iterable[Mapping[str, Any]], credential)
    elif isinstance(credential, (TokenCredential, AsyncTokenCredential)):
        auth['clientSecretCredential'] = credential
    else:
        raise TypeError(
            "Unrecognized credential type. Please supply the master key as a string "
            "or a dictionary, or resource tokens, or a list of permissions, or any instance of a class implementing"
            " AsyncTokenCredential (see azure.identity module for specific implementations "
            "such as ClientSecretCredential).")
    return auth


def _build_connection_policy(kwargs: dict[str, Any]) -> ConnectionPolicy:
    # pylint: disable=protected-access
    policy = kwargs.pop('connection_policy', None) or ConnectionPolicy()

    # Connection config
    # `request_timeout` is supported as a legacy parameter later replaced by `connection_timeout`
    if 'request_timeout' in kwargs:
        policy.RequestTimeout = kwargs.pop('request_timeout') / 1000.0
    else:
        policy.RequestTimeout = kwargs.pop('connection_timeout', policy.RequestTimeout)
    policy.ConnectionMode = kwargs.pop('connection_mode', policy.ConnectionMode)
    policy.ProxyConfiguration = kwargs.pop('proxy_config', policy.ProxyConfiguration)
    policy.EnableEndpointDiscovery = kwargs.pop('enable_endpoint_discovery', policy.EnableEndpointDiscovery)
    policy.PreferredLocations = kwargs.pop('preferred_locations', policy.PreferredLocations)
    policy.ExcludedLocations = kwargs.pop('excluded_locations', policy.ExcludedLocations)
    policy.UseMultipleWriteLocations = kwargs.pop('multiple_write_locations', policy.UseMultipleWriteLocations)

    # SSL config
    verify = kwargs.pop('connection_verify', None)
    policy.DisableSSLVerification = not bool(verify if verify is not None else True)
    ssl = kwargs.pop('ssl_config', None) or policy.SSLConfiguration
    if ssl:
        ssl.SSLCertFile = kwargs.pop('connection_cert', ssl.SSLCertFile)
        ssl.SSLCaCerts = verify or ssl.SSLCaCerts
        policy.SSLConfiguration = ssl

    # Retry config
    retry_options = policy.RetryOptions
    total_retries = kwargs.pop('retry_total', None)
    total_throttle_retries = kwargs.pop('retry_throttle_total', None)
    retry_options._max_retry_attempt_count = \
        total_throttle_retries or total_retries or retry_options._max_retry_attempt_count
    retry_options._fixed_retry_interval_in_milliseconds = \
        kwargs.pop('retry_fixed_interval', retry_options._fixed_retry_interval_in_milliseconds)
    max_backoff = kwargs.pop('retry_backoff_max', None)
    max_throttle_backoff = kwargs.pop('retry_throttle_backoff_max', None)
    retry_options._max_wait_time_in_seconds = \
        max_throttle_backoff or max_backoff or retry_options._max_wait_time_in_seconds
    policy.RetryOptions = retry_options
    connection_retry = policy.ConnectionRetryConfiguration
    if not connection_retry:
        connection_retry = _ConnectionRetryPolicy(
            retry_total=total_retries,
            retry_connect=kwargs.pop('retry_connect', None),
            retry_read=kwargs.pop('retry_read', None),
            retry_status=kwargs.pop('retry_status', None),
            retry_backoff_max=max_backoff or retry_options._max_wait_time_in_seconds,
            retry_mode=kwargs.pop('retry_mode', RetryMode.Fixed),
            retry_on_status_codes=kwargs.pop('retry_on_status_codes', []),
            retry_backoff_factor=kwargs.pop('retry_backoff_factor', 1),
        )
    policy.ConnectionRetryConfiguration = connection_retry
    policy.ResponsePayloadOnWriteDisabled = kwargs.pop('no_response_on_write', False)
    policy.RetryNonIdempotentWrites = kwargs.pop(Constants.Kwargs.RETRY_WRITE, False)
    return policy


class CosmosClient:  # pylint: disable=client-accepts-api-version-keyword
    """A client-side logical representation of an Azure Cosmos DB account.

    Use this client to configure and execute requests to the Azure Cosmos DB service.

    It's recommended to maintain a single instance of CosmosClient per lifetime of the application which enables
        efficient connection management and performance.

    CosmosClient initialization is a heavy operation - don't use initialization CosmosClient instances as
        credentials or network connectivity validations.

    :param str url: The URL of the Cosmos DB account.
    :param credential: Can be the account key, or a dictionary of resource tokens.
    :type credential: Union[str, dict[str, str], ~azure.core.credentials_async.AsyncTokenCredential]
    :keyword str consistency_level: Consistency level to use for the session. Default value is None (account-level).
        More on consistency levels and possible values: https://aka.ms/cosmos-consistency-levels
    :keyword int timeout: An absolute timeout in seconds, for the combined HTTP request and response processing.
    :keyword int connection_timeout: The HTTP request timeout in seconds.
    :keyword str connection_mode: The connection mode for the client - currently only supports 'Gateway'.
    :keyword proxy_config: Connection proxy configuration.
    :paramtype proxy_config: ~azure.cosmos.ProxyConfiguration
    :keyword ssl_config: Connection SSL configuration.
    :paramtype ssl_config: ~azure.cosmos.SSLConfiguration
    :keyword bool connection_verify: Whether to verify the connection, default value is True.
    :keyword str connection_cert: An alternative certificate to verify the connection.
    :keyword int retry_total: Maximum retry attempts.
    :keyword int retry_backoff_max: Maximum retry wait time in seconds.
    :keyword int retry_fixed_interval: Fixed retry interval in milliseconds.
    :keyword int retry_read: Maximum number of socket read retry attempts.
    :keyword int retry_connect: Maximum number of connection error retry attempts.
    :keyword int retry_status: Maximum number of retry attempts on error status codes.
    :keyword list[int] retry_on_status_codes: A list of specific status codes to retry on.
    :keyword float retry_backoff_factor: Factor to calculate wait time between retry attempts.
    :keyword bool retry_write: Indicates whether the SDK should automatically retry write operations for items, even if
        the operation is not guaranteed to be idempotent. This should only be enabled if the application can
        tolerate such risks or has logic to safely detect and handle duplicate operations.
    :keyword bool enable_endpoint_discovery: Enable endpoint discovery for
        geo-replicated database accounts. (Default: True)
    :keyword list[str] preferred_locations: The preferred locations for geo-replicated database accounts.
    :keyword list[str] excluded_locations: The excluded locations to be skipped from preferred locations. The locations
        in this list are specified as the names of the azure Cosmos locations like, 'West US', 'East US' and so on.
        If all preferred locations were excluded, primary/hub location will be used.
    :keyword bool enable_diagnostics_logging: Enable the CosmosHttpLogging policy.
        Must be used along with a logger to work.
    :keyword ~logging.Logger logger: Logger to be used for collecting request diagnostics. Can be passed in at client
        level (to log all requests) or at a single request level. Requests will be logged at INFO level.
    :keyword bool no_response_on_write: Indicates whether service should be instructed to skip sending 
        response payloads for write operations on items by default unless specified differently per operation.
    :keyword int throughput_bucket: The desired throughput bucket for the client
    :keyword str user_agent_suffix: Allows user agent suffix to be specified when creating client

    .. admonition:: Example:

        .. literalinclude:: ../samples/examples_async.py
            :start-after: [START create_client]
            :end-before: [END create_client]
            :language: python
            :dedent: 0
            :caption: Create a new instance of the Cosmos DB client:
            :name: create_client
    """

    def __init__(
            self,
            url: str,
            credential: Union[str, dict[str, str], AsyncTokenCredential],
            *,
            consistency_level: Optional[str] = None,
            **kwargs: Any
    ) -> None:
        """Instantiate a new CosmosClient."""
        auth = _build_auth(credential)
        connection_policy = _build_connection_policy(kwargs)
        self.client_connection = CosmosClientConnection(
            url,
            auth=auth,
            consistency_level=consistency_level,
            connection_policy=connection_policy,
            **kwargs
        )

    def __repr__(self) -> str:
        return "<CosmosClient [{}]>".format(self.client_connection.url_connection)[:1024]

    async def __aenter__(self) -> "CosmosClient":
        await self.client_connection.pipeline_client.__aenter__()
        await self.client_connection._setup()
        return self

    async def __aexit__(self, *args) -> None:
        await self.client_connection._global_endpoint_manager.close() # pylint: disable=protected-access
        return await self.client_connection.pipeline_client.__aexit__(*args)

    async def close(self) -> None:
        """Close this instance of CosmosClient."""
        await self.__aexit__()

    @classmethod
    def from_connection_string(
        cls,
        conn_str: str,
        *,
        credential: Optional[Union[str, dict[str, str]]] = None,
        consistency_level: Optional[str] = None,
        **kwargs: Any
    ) -> "CosmosClient":
        """Create a CosmosClient instance from a connection string.

        This can be retrieved from the Azure portal.For full list of optional
        keyword arguments, see the CosmosClient constructor.

        :param str conn_str: The connection string.
        :keyword credential: Alternative credentials to use instead of the key provided in the connection string.
        :paramtype credential: Union[str, dict[str, str]]
        :keyword str consistency_level: Consistency level to use for the session. Default value is None (account-level).
            More on consistency levels and possible values: https://aka.ms/cosmos-consistency-levels
        :returns: a CosmosClient instance
        :rtype: ~azure.cosmos.aio.CosmosClient
        """
        settings = _parse_connection_str(conn_str, credential)
        return cls(
            url=settings['AccountEndpoint'],
            credential=settings['AccountKey'],
            consistency_level=consistency_level,
            **kwargs
        )

    @overload
    async def create_database(
        self,
        id: str,
        *,
        offer_throughput: Optional[Union[int, ThroughputProperties]] = None,
        initial_headers: Optional[dict[str, str]] = None,
        response_hook: Optional[Callable[[Mapping[str, Any]], None]] = None,
        throughput_bucket: Optional[int] = None,
        return_properties: Literal[False] = False,
        **kwargs: Any
    ) -> DatabaseProxy:
        """
        Create a new database with the given ID (name).

        :param str id: ID (name) of the database to create.
        :keyword Union[int, ~azure.cosmos.ThroughputProperties] offer_throughput: The provisioned throughput
            for this database.
        :keyword dict[str, str] initial_headers: Initial headers to be sent as part of the request.
        :keyword Callable[[dict[str, str], dict[str, Any]], None] response_hook: A callable invoked with
            the response metadata.
        :keyword int throughput_bucket: The desired throughput bucket for the client
        :keyword bool return_properties: Specifies whether to return either a DatabaseProxy
            or a Tuple containing a DatabaseProxy and the associated database properties.
        :raises ~azure.cosmos.exceptions.CosmosResourceExistsError: Database with the given ID already exists.
        :returns: A `DatabaseProxy` instance representing the database.
        :rtype: ~azure.cosmos.aio.DatabaseProxy

        .. admonition:: Example:

            .. literalinclude:: ../samples/examples_async.py
                :start-after: [START create_database]
                :end-before: [END create_database]
                :language: python
                :dedent: 0
                :caption: Create a database in the Cosmos DB account:
                :name: create_database
        """
        ...

    @overload
    async def create_database(
        self,
        id: str,
        *,
        offer_throughput: Optional[Union[int, ThroughputProperties]] = None,
        initial_headers: Optional[dict[str, str]] = None,
        response_hook: Optional[Callable[[Mapping[str, Any]], None]] = None,
        throughput_bucket: Optional[int] = None,
        return_properties: Literal[True],
        **kwargs: Any
    ) -> tuple[DatabaseProxy, CosmosDict]:
        """
        Create a new database with the given ID (name).

        :param str id: ID (name) of the database to create.
        :keyword Union[int, ~azure.cosmos.ThroughputProperties] offer_throughput: The provisioned throughput
            for this database.
        :keyword dict[str, str] initial_headers: Initial headers to be sent as part of the request.
        :keyword Callable[[dict[str, str], dict[str, Any]], None] response_hook: A callable invoked with
            the response metadata.
        :keyword int throughput_bucket: The desired throughput bucket for the client
        :keyword bool return_properties: Specifies whether to return either a DatabaseProxy
            or a Tuple containing a DatabaseProxy and the associated database properties.
        :raises ~azure.cosmos.exceptions.CosmosResourceExistsError: Database with the given ID already exists.
        :returns: A tuple of `DatabaseProxy` and CosmosDict with the database properties.
        :rtype: tuple [~azure.cosmos.aio.DatabaseProxy, ~azure.cosmos.CosmosDict]

        .. admonition:: Example:

            .. literalinclude:: ../samples/examples_async.py
                :start-after: [START create_database]
                :end-before: [END create_database]
                :language: python
                :dedent: 0
                :caption: Create a database in the Cosmos DB account:
                :name: create_database
        """
        ...

    @distributed_trace_async
    async def create_database( # pylint:disable=docstring-should-be-keyword
        self,
        *args: Any,
        **kwargs: Any
    ) -> Union[DatabaseProxy, tuple[DatabaseProxy, CosmosDict]]:
        """
        Create a new database with the given ID (name).

        :param Any args: args
        :param str id: ID (name) of the database to read or create.
        :keyword Union[int, ~azure.cosmos.ThroughputProperties] offer_throughput: The provisioned throughput
            for this database.
        :keyword dict[str, str] initial_headers: Initial headers to be sent as part of the request.
        :keyword Callable[[dict[str, str], dict[str, Any]], None] response_hook: A callable invoked with
            the response metadata.
        :keyword int throughput_bucket: The desired throughput bucket for the client
        :keyword bool return_properties: Specifies whether to return either a DatabaseProxy
            or a Tuple containing a DatabaseProxy and the associated database properties.
        :raises ~azure.cosmos.exceptions.CosmosResourceExistsError: Database with the given ID already exists.
        :returns: A DatabaseProxy instance representing the database or a tuple of DatabaseProxy
            and CosmosDict with the database properties.
        :rtype: ~azure.cosmos.aio.DatabaseProxy or tuple [~azure.cosmos.aio.DatabaseProxy, ~azure.cosmos.CosmosDict]

        .. admonition:: Example:

            .. literalinclude:: ../samples/examples_async.py
                :start-after: [START create_database]
                :end-before: [END create_database]
                :language: python
                :dedent: 0
                :caption: Create a database in the Cosmos DB account:
                :name: create_database
        """
        id = args[0] if args else kwargs.pop("id")
        if len(args) > 1:
            raise TypeError(f"Unexpected positional parameters: {args[1:]}")
        session_token = kwargs.get('session_token')
        if session_token is not None:
            warnings.warn(
                "The 'session_token' flag does not apply to this method and is always ignored even if passed."
                " It will now be removed in the future.",
                DeprecationWarning)
        etag = kwargs.get('etag')
        if etag is not None:
            warnings.warn(
                "The 'etag' flag does not apply to this method and is always ignored even if passed."
                " It will now be removed in the future.",
                DeprecationWarning)
        match_condition = kwargs.get('match_condition')
        if match_condition is not None:
            warnings.warn(
                "The 'match_condition' flag does not apply to this method and is always ignored even if passed."
                " It will now be removed in the future.",
                DeprecationWarning)

        offer_throughput = kwargs.pop("offer_throughput", None)
        return_properties = kwargs.pop("return_properties", False)
        response_hook = kwargs.pop("response_hook", None)

        request_options = _build_options(kwargs)
        _set_throughput_options(offer=offer_throughput, request_options=request_options)

        result = await self.client_connection.CreateDatabase(database={"id": id}, options=request_options, **kwargs)
        if response_hook:
            response_hook(self.client_connection.last_response_headers)
        if not return_properties:
            return DatabaseProxy(self.client_connection, id=result["id"], properties=result)
        return  DatabaseProxy(self.client_connection, id=result["id"], properties=result), result

    @overload
    async def create_database_if_not_exists(  # pylint: disable=redefined-builtin
        self,
        id: str,
        *,
        offer_throughput: Optional[Union[int, ThroughputProperties]] = None,
        initial_headers: Optional[dict[str, str]] = None,
        response_hook: Optional[Callable[[Mapping[str, Any]], None]] = None,
        throughput_bucket: Optional[int] = None,
        return_properties: Literal[False] = False,
        **kwargs: Any
    ) -> DatabaseProxy:
        """
        Create the database if it does not exist already.

        If the database already exists, the existing settings are returned.

        ..note::
            This function does not check or update existing database settings or
            offer throughput if they differ from what is passed in.

        :param str id: ID (name) of the database to read or create.
        :keyword Union[int, ~azure.cosmos.ThroughputProperties] offer_throughput: The provisioned throughput
            for this database.
        :keyword dict[str, str] initial_headers: Initial headers to be sent as part of the request.
        :keyword Callable[[dict[str, str], dict[str, Any]], None] response_hook: A callable invoked with
            the response metadata.
        :keyword int throughput_bucket: The desired throughput bucket for the client
        :keyword bool return_properties: Specifies whether to return either a DatabaseProxy
            or a Tuple containing a DatabaseProxy and the associated database properties.
        :raises ~azure.cosmos.exceptions.CosmosHttpResponseError: The database read or creation failed.
        :returns: A DatabaseProxy instance representing the database.
        :rtype: ~azure.cosmos.aio.DatabaseProxy
        """
        ...

    @overload
    async def create_database_if_not_exists(  # pylint: disable=redefined-builtin
        self,
        id: str,
        *,
        offer_throughput: Optional[Union[int, ThroughputProperties]] = None,
        initial_headers: Optional[dict[str, str]] = None,
        response_hook: Optional[Callable[[Mapping[str, Any]], None]] = None,
        throughput_bucket: Optional[int] = None,
        return_properties: Literal[True],
        **kwargs: Any
    ) -> tuple[DatabaseProxy, CosmosDict]:
        """
        Create the database if it does not exist already.

        If the database already exists, the existing settings are returned.

        ..note::
            This function does not check or update existing database settings or
            offer throughput if they differ from what is passed in.

        :param str id: ID (name) of the database to read or create.
        :keyword Union[int, ~azure.cosmos.ThroughputProperties] offer_throughput: The provisioned throughput
            for this database.
        :keyword dict[str, str] initial_headers: Initial headers to be sent as part of the request.
        :keyword Callable[[dict[str, str], dict[str, Any]], None] response_hook: A callable invoked with
            the response metadata.
        :keyword int throughput_bucket: The desired throughput bucket for the client
        :keyword bool return_properties: Specifies whether to return either a DatabaseProxy
            or a Tuple containing a DatabaseProxy and the associated database properties.
        :raises ~azure.cosmos.exceptions.CosmosHttpResponseError: The database read or creation failed.
        :returns: A tuple of DatabaseProxy and CosmosDict with the database properties.
        :rtype: tuple [~azure.cosmos.aio.DatabaseProxy, ~azure.cosmos.CosmosDict]
        """
        ...

    @distributed_trace_async
    async def create_database_if_not_exists( # pylint:disable=docstring-should-be-keyword
        self,
        *args: Any,
        **kwargs: Any
    ) -> Union[DatabaseProxy, tuple[DatabaseProxy, CosmosDict]]:
        """
        Create the database if it does not exist already.

        If the database already exists, the existing settings are returned.

        ..note::
            This function does not check or update existing database settings or
            offer throughput if they differ from what is passed in.

        :param Any args: args
        :param str id: ID (name) of the database to read or create.
        :keyword Union[int, ~azure.cosmos.ThroughputProperties] offer_throughput: The provisioned throughput
            for this database.
        :keyword dict[str, str] initial_headers: Initial headers to be sent as part of the request.
        :keyword Callable[[dict[str, str], dict[str, Any]], None] response_hook: A callable invoked with
            the response metadata.
        :keyword int throughput_bucket: The desired throughput bucket for the client
        :keyword bool return_properties: Specifies whether to return either a DatabaseProxy
            or a Tuple containing a DatabaseProxy and the associated database properties.
        :raises ~azure.cosmos.exceptions.CosmosHttpResponseError: The database read or creation failed.
        :returns: A DatabaseProxy instance representing the database or a tuple of DatabaseProxy
            and CosmosDict with the database properties.
        :rtype: ~azure.cosmos.aio.DatabaseProxy or tuple [~azure.cosmos.aio.DatabaseProxy, ~azure.cosmos.CosmosDict]
        """

        id = args[0] if args else kwargs.pop("id")
        if len(args) > 1:
            raise TypeError(f"Unexpected positional parameters: {args[1:]}")

        session_token = kwargs.get('session_token')
        if session_token is not None:
            warnings.warn(
                "The 'session_token' flag does not apply to this method and is always ignored even if passed."
                " It will now be removed in the future.",
                DeprecationWarning)
        etag = kwargs.get('etag')
        if etag is not None:
            warnings.warn(
                "The 'etag' flag does not apply to this method and is always ignored even if passed."
                " It will now be removed in the future.",
                DeprecationWarning)
        match_condition = kwargs.get('match_condition')
        if match_condition is not None:
            warnings.warn(
                "The 'match_condition' flag does not apply to this method and is always ignored even if passed."
                " It will now be removed in the future.",
                DeprecationWarning)

        offer_throughput = kwargs.pop("offer_throughput", None)
        return_properties = kwargs.pop("return_properties", False)

        try:
            database_proxy = self.get_database_client(id)
            result = await database_proxy.read(**kwargs)
            if not return_properties:
                return database_proxy
            return database_proxy, result
        except CosmosResourceNotFoundError:
            return await self.create_database(
                id,
                offer_throughput=offer_throughput,
                return_properties=return_properties,
                **kwargs
            )

    def get_database_client(self, database: Union[str, DatabaseProxy, dict[str, Any]]) -> DatabaseProxy:
        """Retrieve an existing database with the ID (name) `id`.

        :param database: The ID (name), dict representing the properties, or :class:`DatabaseProxy`
            instance of the database to get.
        :type database: Union[str, ~azure.cosmos.DatabaseProxy, dict[str, Any]]
        :returns: A `DatabaseProxy` instance representing the retrieved database.
        :rtype: ~azure.cosmos.DatabaseProxy
        """
        if isinstance(database, str):
            id_value = database
        elif isinstance(database, DatabaseProxy):
            id_value = database.id
        else:
            id_value = str(database['id'])
        return DatabaseProxy(self.client_connection, id_value)

    @distributed_trace
    def list_databases(
        self,
        *,
        max_item_count: Optional[int] = None,
        initial_headers: Optional[dict[str, str]] = None,
        response_hook: Optional[Callable[[Mapping[str, Any]], None]] = None,
        throughput_bucket: Optional[int] = None,
        **kwargs: Any
    ) -> AsyncItemPaged[dict[str, Any]]:
        """List the databases in a Cosmos DB SQL database account.

        :keyword int max_item_count: Max number of items to be returned in the enumeration operation.
        :keyword str session_token: Token for use with Session consistency.
        :keyword dict[str, str] initial_headers: Initial headers to be sent as part of the request.
        :keyword response_hook: A callable invoked with the response metadata.
        :keyword int throughput_bucket: The desired throughput bucket for the client
        :paramtype response_hook: Callable[[Mapping[str, Any]], None]
        :returns: An AsyncItemPaged of database properties (dicts).
        :rtype: AsyncItemPaged[dict[str, str]]
        """
        session_token = kwargs.pop('session_token', None)
        if session_token is not None:
            warnings.warn(
                "The 'session_token' flag does not apply to this method and is always ignored even if passed."
                " It will now be removed in the future.",
                DeprecationWarning)
        if initial_headers is not None:
            kwargs["initial_headers"] = initial_headers
        if throughput_bucket is not None:
            kwargs["throughput_bucket"] = throughput_bucket
        feed_options = _build_options(kwargs)
        if max_item_count is not None:
            feed_options["maxItemCount"] = max_item_count

        result = self.client_connection.ReadDatabases(options=feed_options, **kwargs)
        if response_hook:
            response_hook(self.client_connection.last_response_headers)
        return result

    @distributed_trace
    def query_databases(
        self,
        query: str,
        *,
        parameters: Optional[list[dict[str, Any]]] = None,
        max_item_count: Optional[int] = None,
        initial_headers: Optional[dict[str, str]] = None,
        response_hook: Optional[Callable[[Mapping[str, Any]], None]] = None,
        throughput_bucket: Optional[int] = None,
        **kwargs: Any
    ) -> AsyncItemPaged[dict[str, Any]]:
        """Query the databases in a Cosmos DB SQL database account.

        :param Union[str, dict[str, Any]] query: The Azure Cosmos DB SQL query to execute.
        :keyword parameters: Optional array of parameters to the query.
            Each parameter is a dict() with 'name' and 'value' keys.
        :paramtype parameters: list[dict[str, Any]]
        :keyword int max_item_count: Max number of items to be returned in the enumeration operation.
        :keyword str session_token: Token for use with Session consistency.
        :keyword dict[str, str] initial_headers: Initial headers to be sent as part of the request.
        :keyword response_hook: A callable invoked with the response metadata.
        :keyword int throughput_bucket: The desired throughput bucket for the client
        :paramtype response_hook: Callable[[Mapping[str, Any]], None]
        :returns: An AsyncItemPaged of database properties (dicts).
        :rtype: AsyncItemPaged[dict[str, str]]
        """
        session_token = kwargs.get('session_token')
        if session_token is not None:
            warnings.warn(
                "The 'session_token' flag does not apply to this method and is always ignored even if passed."
                " It will now be removed in the future.",
                DeprecationWarning)
        if initial_headers is not None:
            kwargs["initial_headers"] = initial_headers
        if throughput_bucket is not None:
            kwargs['throughput_bucket'] = throughput_bucket
        feed_options = _build_options(kwargs)
        if max_item_count is not None:
            feed_options["maxItemCount"] = max_item_count

        result = self.client_connection.QueryDatabases(
            query=query if parameters is None else {"query": query, "parameters": parameters},
            options=feed_options,
            **kwargs)
        if response_hook:
            response_hook(self.client_connection.last_response_headers)
        return result

    @distributed_trace_async
    async def delete_database(
        self,
        database: Union[str, DatabaseProxy, dict[str, Any]],
        *,
        initial_headers: Optional[dict[str, str]] = None,
        response_hook: Optional[Callable[[Mapping[str, Any]], None]] = None,
        throughput_bucket: Optional[int] = None,
        **kwargs: Any
    ) -> None:
        """Delete the database with the given ID (name).

        :param database: The ID (name), dict representing the properties, or :class:`DatabaseProxy`
            instance of the database to delete.
        :type database: Union[str, ~azure.cosmos.DatabaseProxy, dict[str, Any]]
        :keyword dict[str, str] initial_headers: Initial headers to be sent as part of the request.
        :keyword response_hook: A callable invoked with the response metadata.
        :keyword int throughput_bucket: The desired throughput bucket for the client
        :paramtype response_hook: Callable[[Mapping[str, Any]], None]
        :raises ~azure.cosmos.exceptions.CosmosHttpResponseError: If the database couldn't be deleted.
        :rtype: None
        """
        session_token = kwargs.get('session_token')
        if session_token is not None:
            warnings.warn(
                "The 'session_token' flag does not apply to this method and is always ignored even if passed."
                " It will now be removed in the future.",
                DeprecationWarning)
        etag = kwargs.get('etag')
        if etag is not None:
            warnings.warn(
                "The 'etag' flag does not apply to this method and is always ignored even if passed."
                " It will now be removed in the future.",
                DeprecationWarning)
        match_condition = kwargs.get('match_condition')
        if match_condition is not None:
            warnings.warn(
                "The 'match_condition' flag does not apply to this method and is always ignored even if passed."
                " It will now be removed in the future.",
                DeprecationWarning)
        if throughput_bucket is not None:
            kwargs['throughput_bucket'] = throughput_bucket
        if initial_headers is not None:
            kwargs["initial_headers"] = initial_headers
        request_options = _build_options(kwargs)
        database_link = _get_database_link(database)
        await self.client_connection.DeleteDatabase(database_link, options=request_options, **kwargs)
        if response_hook:
            response_hook(self.client_connection.last_response_headers)

    @distributed_trace_async
    async def _get_database_account(
            self,
            *,
            response_hook: Optional[Callable[[Mapping[str, Any]], None]] = None,
            **kwargs: Any
    ) -> DatabaseAccount:
        """Retrieve the database account information.

        :keyword response_hook: A callable invoked with the response metadata.
        :paramtype response_hook: Callable[[Mapping[str, Any]], None]
        :returns: A `DatabaseAccount` instance representing the Cosmos DB Database Account.
        :rtype: ~azure.cosmos.DatabaseAccount
        """
        result = await self.client_connection.GetDatabaseAccount(**kwargs)
        if response_hook:
            response_hook(self.client_connection.last_response_headers)
        return result



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.

import unittest
import uuid

import pytest

import azure.cosmos._cosmos_client_connection as cosmos_client_connection
import azure.cosmos._global_endpoint_manager as global_endpoint_manager
import azure.cosmos.documents as documents
import azure.cosmos.exceptions as exceptions
import azure.cosmos.http_constants as http_constants
import test_config
from azure.cosmos import _endpoint_discovery_retry_policy
from azure.cosmos import _retry_utility
from azure.cosmos import cosmos_client, PartitionKey
from azure.cosmos._request_object import RequestObject
from azure.cosmos.http_constants import HttpHeaders, StatusCodes, SubStatusCodes


@pytest.mark.cosmosEmulator
@pytest.mark.skip
class TestStreamingFailOver(unittest.TestCase):
    DEFAULT_ENDPOINT = "https://geotest.documents.azure.com:443/"
    MASTER_KEY = "SomeKeyValue"
    WRITE_ENDPOINT1 = "https://geotest-WestUS.documents.azure.com:443/"
    WRITE_ENDPOINT2 = "https://geotest-CentralUS.documents.azure.com:443/"
    READ_ENDPOINT1 = "https://geotest-SouthCentralUS.documents.azure.com:443/"
    READ_ENDPOINT2 = "https://geotest-EastUS.documents.azure.com:443/"
    WRITE_ENDPOINT_NAME1 = "West US"
    WRITE_ENDPOINT_NAME2 = "Central US"
    READ_ENDPOINT_NAME1 = "South Central US"
    READ_ENDPOINT_NAME2 = "East US"
    preferred_regional_endpoints = [READ_ENDPOINT_NAME1, READ_ENDPOINT_NAME2]
    counter = 0
    endpoint_sequence = []

    def test_streaming_fail_over(self):
        self.OriginalExecuteFunction = _retry_utility.ExecuteFunction
        _retry_utility.ExecuteFunction = self._MockExecuteFunctionEndpointDiscover
        connection_policy = documents.ConnectionPolicy()
        connection_policy.PreferredLocations = self.preferred_regional_endpoints
        connection_policy.DisableSSLVerification = True

        client = cosmos_client.CosmosClient(self.DEFAULT_ENDPOINT, self.MASTER_KEY,
                                            consistency_level=documents.ConsistencyLevel.Eventual,
                                            connection_policy=connection_policy)
        self.original_get_database_account = client.client_connection.GetDatabaseAccount
        self.original_get_read_endpoints = (client.client_connection._global_endpoint_manager.location_cache
                                            .get_read_regional_routing_contexts())
        self.original_get_write_endpoints = (client.client_connection._global_endpoint_manager.location_cache
                                             .get_write_regional_routing_contexts())
        client.client_connection.GetDatabaseAccount = self.mock_get_database_account
        client.client_connection._global_endpoint_manager.location_cache.get_read_regional_routing_contexts = (
            self.mock_get_read_endpoints)
        client.client_connection._global_endpoint_manager.location_cache.get_write_regional_routing_contexts = (
            self.mock_get_write_endpoints)
        created_db = client.create_database_if_not_exists("streaming-db" + str(uuid.uuid4()))
        created_container = created_db.create_container("streaming-container" + str(uuid.uuid4()),
                                                        PartitionKey(path="/id"))

        document_definition = {'id': 'doc',
                               'name': 'sample document',
                               'key': 'value'}

        created_document = created_container.create_item(document_definition)

        self.assertDictEqual(created_document, {})
        self.assertDictEqual(created_document.get_response_headers(), {})

        self.assertEqual(self.counter, 10)
        # First request is an initial read collection.
        # Next 6 requests hit forbidden write exceptions and the endpoint retry policy keeps
        # flipping the resolved endpoint between the 2 write endpoints.
        # The 10th request returns the actual read document.
        for i in range(0, 6):
            if i % 2 == 0:
                self.assertEqual(self.endpoint_sequence[i], self.WRITE_ENDPOINT1)
            else:
                self.assertEqual(self.endpoint_sequence[i], self.WRITE_ENDPOINT2)

        cosmos_client_connection.CosmosClientConnection.GetDatabaseAccount = self.original_get_database_account
        _retry_utility.ExecuteFunction = self.OriginalExecuteFunction
        client.client_connection._global_endpoint_manager.location_cache.get_read_regional_routing_contexts = (
            self.original_get_read_endpoints)
        client.client_connection._global_endpoint_manager.location_cache.get_write_regional_routing_contexts = (
            self.original_get_write_endpoints)

    def mock_get_database_account(self, url_connection=None):
        database_account = documents.DatabaseAccount()
        database_account._EnableMultipleWritableLocations = True
        database_account._WritableLocations = [
            {'name': self.WRITE_ENDPOINT_NAME1, 'databaseAccountEndpoint': self.WRITE_ENDPOINT1},
            {'name': self.WRITE_ENDPOINT_NAME2, 'databaseAccountEndpoint': self.WRITE_ENDPOINT2}
        ]
        database_account._ReadableLocations = [
            {'name': self.READ_ENDPOINT_NAME1, 'databaseAccountEndpoint': self.READ_ENDPOINT1},
            {'name': self.READ_ENDPOINT_NAME2, 'databaseAccountEndpoint': self.READ_ENDPOINT2}
        ]
        return database_account

    def mock_get_read_endpoints(self):
        return [
            {'name': self.READ_ENDPOINT_NAME1, 'databaseAccountEndpoint': self.READ_ENDPOINT1},
            {'name': self.READ_ENDPOINT_NAME2, 'databaseAccountEndpoint': self.READ_ENDPOINT2}
        ]

    def mock_get_write_endpoints(self):
        return [
            {'name': self.WRITE_ENDPOINT_NAME1, 'databaseAccountEndpoint': self.WRITE_ENDPOINT1},
            {'name': self.WRITE_ENDPOINT_NAME2, 'databaseAccountEndpoint': self.WRITE_ENDPOINT2}
        ]

    def _MockExecuteFunctionEndpointDiscover(self, function, *args, **kwargs):
        self.counter += 1
        if self.counter >= 10 or (len(args) > 0 and args[1].operation_type == documents._OperationType.Read):
            return {}, {}
        else:
            self.endpoint_sequence.append(args[1].location_endpoint_to_route)
            response = test_config.FakeResponse({HttpHeaders.SubStatus: SubStatusCodes.WRITE_FORBIDDEN})
            raise exceptions.CosmosHttpResponseError(
                status_code=StatusCodes.FORBIDDEN,
                message="Request is not permitted in this region",
                response=response)

    def test_retry_policy_does_not_mark_null_locations_unavailable(self):
        self.OriginalExecuteFunction = _retry_utility.ExecuteFunction
        _retry_utility.ExecuteFunction = self._MockExecuteFunctionEndpointDiscover
        connection_policy = documents.ConnectionPolicy()
        connection_policy.PreferredLocations = self.preferred_regional_endpoints
        connection_policy.DisableSSLVerification = True

        client = cosmos_client.CosmosClient(self.DEFAULT_ENDPOINT, self.MASTER_KEY,
                                            consistency_level=documents.ConsistencyLevel.Eventual,
                                            connection_policy=connection_policy)
        self.original_get_database_account = client.client_connection.GetDatabaseAccount
        client.client_connection.GetDatabaseAccount = self.mock_get_database_account

        endpoint_manager = global_endpoint_manager._GlobalEndpointManager(client.client_connection)

        self.original_mark_endpoint_unavailable_for_read_function = endpoint_manager.mark_endpoint_unavailable_for_read
        endpoint_manager.mark_endpoint_unavailable_for_read = self._mock_mark_endpoint_unavailable_for_read
        self.original_mark_endpoint_unavailable_for_write_function = endpoint_manager.mark_endpoint_unavailable_for_write
        endpoint_manager.mark_endpoint_unavailable_for_write = self._mock_mark_endpoint_unavailable_for_write
        self.original_resolve_service_endpoint = endpoint_manager.resolve_service_endpoint_for_partition
        endpoint_manager.resolve_service_endpoint_for_partition = self._mock_resolve_service_endpoint

        # Read and write counters count the number of times the endpoint manager's
        # mark_endpoint_unavailable_for_read() and mark_endpoint_unavailable_for_read() 
        # functions were called. When a 'None' location is returned by resolve_service_endpoint(),
        # these functions  should not be called
        self._read_counter = 0
        self._write_counter = 0
        request = RequestObject(http_constants.ResourceType.Document, documents._OperationType.Read)
        endpoint_discovery_retry_policy = _endpoint_discovery_retry_policy.EndpointDiscoveryRetryPolicy(
            documents.ConnectionPolicy(), endpoint_manager, request)
        endpoint_discovery_retry_policy.ShouldRetry(exceptions.CosmosHttpResponseError(
            status_code=http_constants.StatusCodes.FORBIDDEN))
        self.assertEqual(self._read_counter, 0)
        self.assertEqual(self._write_counter, 0)

        self._read_counter = 0
        self._write_counter = 0
        request = RequestObject(http_constants.ResourceType.Document, documents._OperationType.Create)
        endpoint_discovery_retry_policy = _endpoint_discovery_retry_policy.EndpointDiscoveryRetryPolicy(
            documents.ConnectionPolicy(), endpoint_manager, request)
        endpoint_discovery_retry_policy.ShouldRetry(exceptions.CosmosHttpResponseError(
            status_code=http_constants.StatusCodes.FORBIDDEN))
        self.assertEqual(self._read_counter, 0)
        self.assertEqual(self._write_counter, 0)

        endpoint_manager.mark_endpoint_unavailable_for_read = (self
                                                               .original_mark_endpoint_unavailable_for_read_function)
        endpoint_manager.mark_endpoint_unavailable_for_write = (self.
                                                                original_mark_endpoint_unavailable_for_write_function)
        cosmos_client_connection.CosmosClientConnection.GetDatabaseAccount = self.original_get_database_account

    def _mock_mark_endpoint_unavailable_for_read(self, endpoint):
        self._read_counter += 1
        self.original_mark_endpoint_unavailable_for_read_function(endpoint)

    def _mock_mark_endpoint_unavailable_for_write(self, endpoint):
        self._write_counter += 1
        self.original_mark_endpoint_unavailable_for_write_function(endpoint)

    @staticmethod
    def _mock_resolve_service_endpoint(request):
        return None


if __name__ == '__main__':
    unittest.main()



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.

import unittest
import uuid

import pytest

import test_config
from azure.core import MatchConditions
from azure.cosmos import Offer, http_constants, CosmosClient, DatabaseProxy, ContainerProxy, PartitionKey
from azure.cosmos.exceptions import CosmosHttpResponseError

def check_pk_range_statistics_request_headers(raw_response):
    assert raw_response.http_request.headers[http_constants.HttpHeaders.PopulatePartitionKeyRangeStatistics] == 'True'

def check_quota_info_request_headers(raw_response):
    assert raw_response.http_request.headers[http_constants.HttpHeaders.PopulateQuotaInfo] == 'True'

@pytest.mark.cosmosEmulator
class TestBackwardsCompatibility(unittest.TestCase):
    configs = test_config.TestConfig
    databaseForTest: DatabaseProxy = None
    client: CosmosClient = None
    containerForTest: ContainerProxy = None
    host = configs.host
    masterKey = configs.masterKey

    populate_true = True

    @classmethod
    def setUpClass(cls):
        if cls.masterKey == '[YOUR_KEY_HERE]' or cls.host == '[YOUR_ENDPOINT_HERE]':
            raise Exception(
                "You must specify your Azure Cosmos account values for "
                "'masterKey' and 'host' at the top of this class to run the "
                "tests.")
        cls.client = CosmosClient(cls.host, cls.masterKey)
        cls.databaseForTest = cls.client.get_database_client(cls.configs.TEST_DATABASE_ID)
        cls.containerForTest = cls.databaseForTest.get_container_client(cls.configs.TEST_SINGLE_PARTITION_CONTAINER_ID)

    def test_offer_methods(self):
        database_offer = self.databaseForTest.get_throughput()
        container_offer = self.containerForTest.get_throughput()

        self.assertTrue("ThroughputProperties" in str(type(database_offer)))
        self.assertTrue("ThroughputProperties" in str(type(container_offer)))

        self.assertTrue(isinstance(database_offer, Offer))
        self.assertTrue(isinstance(container_offer, Offer))

    def test_populate_quota_info(self):
            self.containerForTest.read(populate_quota_info=True, raw_response_hook=check_quota_info_request_headers)
            self.containerForTest.read(False, False, True, raw_response_hook=check_quota_info_request_headers)

    def test_populate_partition_key_range_statistics(self):
        self.containerForTest.read(populate_partition_key_range_statistics=True, raw_response_hook=check_pk_range_statistics_request_headers)
        self.containerForTest.read(False, True, raw_response_hook=check_pk_range_statistics_request_headers)

    def test_session_token_compatibility(self):
        # Verifying that behavior is unaffected across the board for using `session_token` on irrelevant methods
        # Database
        database = self.client.create_database(str(uuid.uuid4()), session_token=str(uuid.uuid4()))
        assert database is not None
        database2 = self.client.create_database_if_not_exists(str(uuid.uuid4()), session_token=str(uuid.uuid4()))
        assert database2 is not None
        database_list = list(self.client.list_databases(session_token=str(uuid.uuid4())))
        database_list2 = list(self.client.query_databases(query="select * from c", session_token=str(uuid.uuid4())))
        assert len(database_list) > 0
        assert len(database_list2) > 0
        database_read = database.read(session_token=str(uuid.uuid4()))
        assert database_read is not None
        self.client.delete_database(database2.id, session_token=str(uuid.uuid4()))
        try:
            database2.read()
            pytest.fail("Database read should have failed")
        except CosmosHttpResponseError as e:
            assert e.status_code == 404

        # Container
        container = self.databaseForTest.create_container(str(uuid.uuid4()), PartitionKey(path="/pk"), session_token=str(uuid.uuid4()))
        assert container is not None
        container2 = self.databaseForTest.create_container_if_not_exists(str(uuid.uuid4()), PartitionKey(path="/pk"), session_token=str(uuid.uuid4()))
        assert container2 is not None
        container_list = list(self.databaseForTest.list_containers(session_token=str(uuid.uuid4())))
        container_list2 = list(self.databaseForTest.query_containers(query="select * from c", session_token=str(uuid.uuid4())))
        assert len(container_list) > 0
        assert len(container_list2) > 0
        container2_read = container2.read(session_token=str(uuid.uuid4()))
        assert container2_read is not None
        replace_container = self.databaseForTest.replace_container(container2, PartitionKey(path="/pk"), default_ttl=30, session_token=str(uuid.uuid4()))
        replace_container_read = replace_container.read()
        assert replace_container is not None
        assert replace_container_read != container2_read
        assert 'defaultTtl' in replace_container_read # Check for default_ttl as a new additional property
        self.databaseForTest.delete_container(replace_container.id, session_token=str(uuid.uuid4()))
        try:
            container2.read()
            pytest.fail("Container read should have failed")
        except CosmosHttpResponseError as e:
            assert e.status_code == 404

        self.client.delete_database(database.id)

    def test_etag_match_condition_compatibility(self):
        # Verifying that behavior is unaffected across the board for using `etag`/`match_condition` on irrelevant methods
        # Database
        database = self.client.create_database(str(uuid.uuid4()), etag=str(uuid.uuid4()), match_condition=MatchConditions.IfModified)
        assert database is not None
        database2 = self.client.create_database_if_not_exists(str(uuid.uuid4()), etag=str(uuid.uuid4()), match_condition=MatchConditions.IfNotModified)
        assert database2 is not None
        self.client.delete_database(database2.id, etag=str(uuid.uuid4()), match_condition=MatchConditions.IfModified)
        try:
            database2.read()
            pytest.fail("Database read should have failed")
        except CosmosHttpResponseError as e:
            assert e.status_code == 404

        # Container
        container = self.databaseForTest.create_container(str(uuid.uuid4()), PartitionKey(path="/pk"),
                                              etag=str(uuid.uuid4()), match_condition=MatchConditions.IfModified)
        assert container is not None
        container2 = self.databaseForTest.create_container_if_not_exists(str(uuid.uuid4()), PartitionKey(path="/pk"),
                                                             etag=str(uuid.uuid4()), match_condition=MatchConditions.IfNotModified)
        assert container2 is not None
        container2_read = container2.read()
        assert container2_read is not None
        replace_container = self.databaseForTest.replace_container(container2, PartitionKey(path="/pk"), default_ttl=30,
                                                       etag=str(uuid.uuid4()), match_condition=MatchConditions.IfModified)
        replace_container_read = replace_container.read()
        assert replace_container is not None
        assert replace_container_read != container2_read
        assert 'defaultTtl' in replace_container_read # Check for default_ttl as a new additional property
        self.databaseForTest.delete_container(replace_container.id, etag=str(uuid.uuid4()), match_condition=MatchConditions.IfModified)
        try:
            container2.read()
            pytest.fail("Container read should have failed")
        except CosmosHttpResponseError as e:
            assert e.status_code == 404

        # Item
        item = container.create_item({"id": str(uuid.uuid4()), "pk": 0}, etag=str(uuid.uuid4()), match_condition=MatchConditions.IfModified)
        assert item is not None
        item2 = container.upsert_item({"id": str(uuid.uuid4()), "pk": 0}, etag=str(uuid.uuid4()),
                                     match_condition=MatchConditions.IfNotModified)
        assert item2 is not None
        item = container.create_item({"id": str(uuid.uuid4()), "pk": 0}, etag=None, match_condition=None)
        assert item is not None
        item2 = container.upsert_item({"id": str(uuid.uuid4()), "pk": 0}, etag=None, match_condition=None)
        assert item2 is not None
        batch_operations = [
            ("create", ({"id": str(uuid.uuid4()), "pk": 0},)),
            ("replace", (item2['id'], {"id": str(uuid.uuid4()), "pk": 0})),
            ("read", (item['id'],)),
            ("upsert", ({"id": str(uuid.uuid4()), "pk": 0},)),
        ]
        batch_results = container.execute_item_batch(batch_operations, partition_key=0, etag=str(uuid.uuid4()), match_condition=MatchConditions.IfModified)
        assert len(batch_results) == 4
        for result in batch_results:
            assert result['statusCode'] in (200, 201)

        self.client.delete_database(database.id)

if __name__ == "__main__":
    unittest.main()



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.

import unittest
import uuid
from datetime import datetime, timedelta, timezone

import pytest
import pytest_asyncio

import test_config
from azure.cosmos.aio import CosmosClient
from azure.cosmos.partition_key import PartitionKey

ID = 'id'
CURRENT = 'current'
PREVIOUS = 'previous'
METADATA = 'metadata'
OPERATION_TYPE = 'operationType'
CREATE = 'create'
DELETE = 'delete'
E_TAG = 'etag'

@pytest_asyncio.fixture()
async def setup():
    config = test_config.TestConfig()
    if config.masterKey == '[YOUR_KEY_HERE]' or config.host == '[YOUR_ENDPOINT_HERE]':
        raise Exception(
            "You must specify your Azure Cosmos account values for "
            "'masterKey' and 'host' at the top of this class to run the "
            "tests.")
    test_client = CosmosClient(config.host, config.masterKey)
    created_db = await test_client.create_database_if_not_exists(config.TEST_DATABASE_ID)
    created_db_data = {
        "created_db": created_db,
        "is_emulator": config.is_emulator
    }

    yield created_db_data
    await test_client.close()

def round_time():
    utc_now = datetime.now(timezone.utc)
    return utc_now - timedelta(microseconds=utc_now.microsecond)

async def assert_change_feed(expected, actual):
    if len(actual) == 0:
        assert len(expected) == len(actual)
        return

    #TODO: remove this if we can add flag to get 'previous' always
    for item in actual:
        if METADATA in item and item[METADATA][OPERATION_TYPE] == DELETE:
            if ID in item[METADATA]:
                item[PREVIOUS] = {ID: item[METADATA][ID]}

    # Sort actual by operation_type and id
    actual = sorted(actual, key=lambda k: (k[METADATA][OPERATION_TYPE], k[CURRENT][ID]) if k[METADATA][OPERATION_TYPE] == CREATE else (k[METADATA][OPERATION_TYPE], k[PREVIOUS][ID]))

    for expected_change_feed, actual_change_feed in zip(expected, actual):
        for expected_type, expected_data in expected_change_feed.items():
            assert expected_type in actual_change_feed
            actual_data = actual_change_feed[expected_type]
            for key, value in expected_data.items():
                assert key in actual_data
                assert expected_data[key] == actual_data[key]

@pytest.mark.cosmosEmulator
@pytest.mark.asyncio
@pytest.mark.usefixtures("setup")
class TestAllVersionsChangeFeedAsync:
    """Test to verify All Versions And Delete change feed behavior"""

    async def test_query_change_feed_all_versions_and_deletes_async(self, setup):
        partition_key = 'pk'
        # 'retentionDuration' was required to enable `ALL_VERSIONS_AND_DELETES` for Emulator testing
        change_feed_policy = {"retentionDuration": 10} if setup["is_emulator"] else None
        created_collection = await setup["created_db"].create_container("change_feed_test_" + str(uuid.uuid4()),
                                                              PartitionKey(path=f"/{partition_key}"),
                                                              change_feed_policy=change_feed_policy)

        mode = 'AllVersionsAndDeletes'

        ## Test Change Feed with empty collection(Save the continuation token)
        query_iterable = created_collection.query_items_change_feed(
            mode=mode,
        )
        expected_change_feeds = []
        actual_change_feeds = [item async for item in query_iterable]
        cont_token1 = created_collection.client_connection.last_response_headers[E_TAG]
        await assert_change_feed(expected_change_feeds, actual_change_feeds)

        ## Test change_feed for created items from cont_token1 (Save the new continuation token)
        new_documents = [{partition_key: f'pk{i}', ID: f'doc{i}'} for i in range(4)]
        created_items = []
        for document in new_documents:
            created_item = await created_collection.create_item(body=document)
            created_items.append(created_item)
        query_iterable = created_collection.query_items_change_feed(
            continuation=cont_token1,
            mode=mode,
        )

        expected_change_feeds = [{CURRENT: {ID: f'doc{i}'}, METADATA: {OPERATION_TYPE: CREATE}} for i in range(4)]
        actual_change_feeds = [item async for item in query_iterable]
        cont_token2 = created_collection.client_connection.last_response_headers['etag']
        await assert_change_feed(expected_change_feeds, actual_change_feeds)

        ## Test change_feed for deleted items
        for item in created_items:
            await created_collection.delete_item(item=item, partition_key=item['pk'])
        query_iterable = created_collection.query_items_change_feed(
            continuation=cont_token2,
            mode=mode,
        )

        expected_change_feeds = [{CURRENT: {}, PREVIOUS: {ID: f'doc{i}'}, METADATA: {OPERATION_TYPE: DELETE}} for i in
                                 range(4)]
        actual_change_feeds = [item async for item in query_iterable]
        await assert_change_feed(expected_change_feeds, actual_change_feeds)

        ## Test change_feed for created/deleted items
        query_iterable = created_collection.query_items_change_feed(
            continuation=cont_token1,
            mode=mode
        )

        expected_change_feeds = [{CURRENT: {ID: f'doc{i}'}, METADATA: {OPERATION_TYPE: CREATE}} for i in range(4)] \
                                + [{CURRENT: {}, PREVIOUS: {ID: f'doc{i}'}, METADATA: {OPERATION_TYPE: DELETE}} for i in
                                   range(4)]
        actual_change_feeds = [item async for item in query_iterable]
        await assert_change_feed(expected_change_feeds, actual_change_feeds)

        query_iterable = created_collection.query_items_change_feed(
            mode = mode,
            partition_key = 'pk1'
        )
        empty_list = [item async for item in query_iterable]
        pk_cont_token = created_collection.client_connection.last_response_headers[E_TAG]
        new_documents = [{partition_key: f'pk{i}', ID: f'doc{i}'} for i in range(4)]
        created_items = []
        for document in new_documents:
            created_item = await created_collection.create_item(body=document)
            created_items.append(created_item)
        query_iterable = created_collection.query_items_change_feed(
            continuation=pk_cont_token,
            mode=mode
        )
        expected_change_feeds = [{CURRENT: {ID: f'doc1'}, METADATA: {OPERATION_TYPE: CREATE}}]
        actual_change_feeds = [item async for item in query_iterable]
        await assert_change_feed(expected_change_feeds, actual_change_feeds)

    async def test_query_change_feed_all_versions_and_deletes_errors_async(self, setup):
        created_collection = await setup["created_db"].create_container("change_feed_test_" + str(uuid.uuid4()),
                                                                  PartitionKey(path="/pk"))
        mode = 'AllVersionsAndDeletes'

        # Error if invalid mode was used
        with pytest.raises(ValueError) as e:
            created_collection.query_items_change_feed(
                mode="test_invalid_mode",
            )
        assert str(e.value) == "Invalid mode was used: 'test_invalid_mode'. Supported modes are ['LatestVersion', 'AllVersionsAndDeletes']."

        # Error if partition_key_range_id was used with FULL_FIDELITY_FEED
        with pytest.raises(ValueError) as e:
            created_collection.query_items_change_feed(
                partition_key_range_id="TestPartitionKeyRangeId",
                mode=mode,
            )
        assert str(e.value) == "'AllVersionsAndDeletes' mode is not supported if 'partition_key_range_id' was used. Please use 'feed_range' instead."

        # Error if is_start_from_beginning was in invalid type
        with pytest.raises(TypeError) as e:
            created_collection.query_items_change_feed(
                is_start_from_beginning="Now",
            )
        assert str(e.value) == "'is_start_from_beginning' must be 'bool' type, but given 'str'."

        # Error if is_start_from_beginning was used with FULL_FIDELITY_FEED
        with pytest.raises(ValueError) as e:
            created_collection.query_items_change_feed(
                is_start_from_beginning="Now",
                mode=mode,
            )
        assert str(e.value) == "'AllVersionsAndDeletes' mode is only supported if 'is_start_from_beginning' is 'False'. Please use 'is_start_from_beginning=False' or 'continuation' instead."

        # Error if 'is_start_from_beginning' was used with 'start_time'
        with pytest.raises(ValueError) as e:
            created_collection.query_items_change_feed(
                is_start_from_beginning=True,
                start_time="Now",
            )
        assert str(e.value) == "'is_start_from_beginning' and 'start_time' are exclusive, please only set one of them."

        # Error if 'start_time' was invalid value
        invalid_time = "Invalid value"
        # TODO: previously it is throwing AttributeError, now has changed into ValueError, is it breaking change?
        with pytest.raises(ValueError) as e:
            created_collection.query_items_change_feed(start_time=invalid_time)
        assert str(e.value) == "'start_time' must be either 'Now' or 'Beginning', but given 'Invalid value'."

        # Error if 'start_time' was invalid type
        invalid_time = 1.2
        with pytest.raises(AttributeError) as e:
            created_collection.query_items_change_feed(start_time=invalid_time)
        assert str(e.value) == "'float' object has no attribute 'lower'"

        # Error if start_time was used with FULL_FIDELITY_FEED
        with pytest.raises(ValueError) as e:
            created_collection.query_items_change_feed(
                start_time=round_time(),
                mode=mode,
            )
        assert str(e.value) == "'AllVersionsAndDeletes' mode is only supported if 'start_time' is 'Now'. Please use 'start_time=\"Now\"' or 'continuation' instead."

if __name__ == '__main__':
    unittest.main()



# The MIT License (MIT)
# Copyright (c) 2014 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Base functions in the Azure Cosmos database service.
"""

import base64
import time
from email.utils import formatdate
import json
import uuid
import re
import binascii
from typing import Any, Mapping, Optional, Sequence, Union, Tuple, TYPE_CHECKING

from urllib.parse import quote as urllib_quote
from urllib.parse import unquote as urllib_unquote
from urllib.parse import urlsplit
from azure.core import MatchConditions

from . import documents
from . import http_constants
from . import _runtime_constants
from ._constants import _Constants as Constants
from .auth import _get_authorization_header
from .offer import ThroughputProperties
from .partition_key import _Empty, _Undefined

if TYPE_CHECKING:
    from ._cosmos_client_connection import CosmosClientConnection
    from .aio._cosmos_client_connection_async import CosmosClientConnection as AsyncClientConnection
    from ._request_object import RequestObject

# pylint: disable=protected-access

_COMMON_OPTIONS = {
    'initial_headers': 'initialHeaders',
    'pre_trigger_include': 'preTriggerInclude',
    'post_trigger_include': 'postTriggerInclude',
    'access_condition': 'accessCondition',
    'session_token': 'sessionToken',
    'resource_token_expiry_seconds': 'resourceTokenExpirySeconds',
    'offer_enable_ru_per_minute_throughput': 'offerEnableRUPerMinuteThroughput',
    'disable_ru_per_minute_usage': 'disableRUPerMinuteUsage',
    'continuation': 'continuation',
    'content_type': 'contentType',
    'is_query_plan_request': 'isQueryPlanRequest',
    'supported_query_features': 'supportedQueryFeatures',
    'query_version': 'queryVersion',
    'priority': 'priorityLevel',
    'no_response': 'responsePayloadOnWriteDisabled',
    'retry_write': Constants.Kwargs.RETRY_WRITE,
    'max_item_count': 'maxItemCount',
    'throughput_bucket': 'throughputBucket',
    'excluded_locations': Constants.Kwargs.EXCLUDED_LOCATIONS
}

# Cosmos resource ID validation regex breakdown:
# ^ Match start of string.
# [^/\#?] Match any character that is not /\#?\n\r\t.
# $ End of string
_VALID_COSMOS_RESOURCE = re.compile(r"^[^/\\#?\t\r\n]*$")


def _get_match_headers(kwargs: dict[str, Any]) -> Tuple[Optional[str], Optional[str]]:
    if_match = kwargs.pop('if_match', None)
    if_none_match = kwargs.pop('if_none_match', None)
    match_condition = kwargs.pop('match_condition', None)
    if match_condition == MatchConditions.IfNotModified:
        if_match = kwargs.pop('etag', None)
        if not if_match:
            raise ValueError("'match_condition' specified without 'etag'.")
    elif match_condition == MatchConditions.IfPresent:
        if_match = '*'
    elif match_condition == MatchConditions.IfModified:
        if_none_match = kwargs.pop('etag', None)
        if not if_none_match:
            raise ValueError("'match_condition' specified without 'etag'.")
    elif match_condition == MatchConditions.IfMissing:
        if_none_match = '*'
    elif match_condition is None:
        etag = kwargs.pop('etag', None)
        if etag is not None:
            raise ValueError("'etag' specified without 'match_condition'.")
    else:
        raise TypeError("Invalid match condition: {}".format(match_condition))
    return if_match, if_none_match


def build_options(kwargs: dict[str, Any]) -> dict[str, Any]:
    options = kwargs.pop('request_options', kwargs.pop('feed_options', {}))
    for key, value in _COMMON_OPTIONS.items():
        if key in kwargs:
            options[value] = kwargs.pop(key)
    if 'read_timeout' in kwargs:
        options['read_timeout'] = kwargs['read_timeout']
    if 'timeout' in kwargs:
        options['timeout'] = kwargs['timeout']


    options[Constants.OperationStartTime] = time.time()
    if_match, if_none_match = _get_match_headers(kwargs)
    if if_match:
        options['accessCondition'] = {'type': 'IfMatch', 'condition': if_match}
    if if_none_match:
        options['accessCondition'] = {'type': 'IfNoneMatch', 'condition': if_none_match}
    return options

def _merge_query_results(
        results: dict[str, Any],
        partial_result: dict[str, Any],
        query: Optional[Union[str, dict[str, Any]]]
) -> dict[str, Any]:
    """Merges partial query results from different partitions.

    This method is required for queries that are manually fanned out to multiple
    partitions or ranges within the SDK, such as prefix partition key queries.
    For non-aggregated queries, results from each partition are simply concatenated.
    However, for aggregate queries (COUNT, SUM, MIN, MAX, AVG), each partition
    returns a partial aggregate. This method merges these partial results to compute
    the final, correct aggregate value.

    TODO:This client-side aggregation is a temporary workaround. Ideally, this logic
    should be integrated into the core pipeline as aggregate queries are handled by DefaultExecutionContext,
    not MultiAggregatorExecutionContext, which is not split proof until the logic is moved to the core pipeline.
    This method handles the aggregation of results when a query spans multiple
    partitions. It specifically handles:
    1. Standard queries: Appends documents from partial_result to results.
    2. Aggregate queries that return a JSON object (e.g., `SELECT COUNT(1) FROM c`, `SELECT MIN(c.field) FROM c`).
    3. VALUE queries with aggregation that return a scalar value (e.g., `SELECT VALUE COUNT(1) FROM c`).

    :param dict[str, Any] results: The accumulated result's dictionary.
    :param dict[str, Any] partial_result: The new partial result dictionary to merge.
    :param query: The query being executed.
    :type query: str or dict[str, Any]
    :return: The merged result's dictionary.
    :rtype: dict[str, Any]
    """
    if not results:
        return partial_result

    partial_docs = partial_result.get("Documents")
    if not partial_docs:
        return results

    results_docs = results.get("Documents")

    # Check if both results are aggregate queries
    is_partial_agg = (
            isinstance(partial_docs, list)
            and len(partial_docs) == 1
            and isinstance(partial_docs[0], dict)
            and partial_docs[0].get("_aggregate") is not None
    )
    is_results_agg = (
            results_docs
            and isinstance(results_docs, list)
            and len(results_docs) == 1
            and isinstance(results_docs[0], dict)
            and results_docs[0].get("_aggregate") is not None
    )

    if is_partial_agg and is_results_agg:
        agg_results = results_docs[0]["_aggregate"] # type: ignore[index]
        agg_partial = partial_docs[0]["_aggregate"]
        for key in agg_partial:
            if key not in agg_results:
                agg_results[key] = agg_partial[key]
            elif isinstance(agg_partial.get(key), dict) and "count" in agg_partial[key]:  # AVG
                if isinstance(agg_results.get(key), dict):
                    agg_results[key]["sum"] += agg_partial[key]["sum"]
                    agg_results[key]["count"] += agg_partial[key]["count"]
            elif key.lower().startswith("min"):
                agg_results[key] = min(agg_results[key], agg_partial[key])
            elif key.lower().startswith("max"):
                agg_results[key] = max(agg_results[key], agg_partial[key])
            else:  # COUNT, SUM
                agg_results[key] += agg_partial[key]
        return results

    # Check if both are VALUE aggregate queries
    is_partial_value_agg = (
            isinstance(partial_docs, list)
            and len(partial_docs) == 1
            and isinstance(partial_docs[0], (int, float))
    )
    is_results_value_agg = (
            results_docs
            and isinstance(results_docs, list)
            and len(results_docs) == 1
            and isinstance(results_docs[0], (int, float))
    )

    if is_partial_value_agg and is_results_value_agg:
        query_text = query.get("query") if isinstance(query, dict) else query
        if query_text:
            query_upper = query_text.upper()
            # For MIN/MAX, we find the min/max of the partial results.
            # For COUNT/SUM, we sum the partial results.
            # Without robust query parsing, we can't distinguish them reliably.
            # Defaulting to sum for COUNT/SUM. MIN/MAX VALUE queries are not fully supported client-side.
            if " SELECT VALUE MIN" in query_upper:
                results_docs[0] = min(results_docs[0], partial_docs[0]) # type: ignore[index]
            elif " SELECT VALUE MAX" in query_upper:
                results_docs[0] = max(results_docs[0], partial_docs[0]) # type: ignore[index]
            else:  # For COUNT/SUM, we sum the partial results
                results_docs[0] += partial_docs[0] # type: ignore[index]
        return results

    # Standard query, append documents
    if results_docs is None:
        results["Documents"] = partial_docs
    elif isinstance(results_docs, list) and isinstance(partial_docs, list):
        results_docs.extend(partial_docs)
    results["_count"] = len(results["Documents"])
    return results


def GetHeaders(  # pylint: disable=too-many-statements,too-many-branches
        cosmos_client_connection: Union["CosmosClientConnection", "AsyncClientConnection"],
        default_headers: Mapping[str, Any],
        verb: str,
        path: str,
        resource_id: Optional[str],
        resource_type: str,
        operation_type: str,
        options: Mapping[str, Any],
        partition_key_range_id: Optional[str] = None,
        client_id: Optional[str] = None,
) -> dict[str, Any]:
    """Gets HTTP request headers.

    :param _cosmos_client_connection.CosmosClientConnection cosmos_client_connection:
    :param dict default_headers:
    :param str verb:
    :param str path:
    :param str resource_id:
    :param str resource_type:
    :param str operation_type:
    :param dict options:
    :param str partition_key_range_id:
    :param str client_id:
    :return: The HTTP request headers.
    :rtype: dict
    """
    headers = dict(default_headers)
    options = options or {}

    # SDK supported capabilities header for partition merge support
    headers[http_constants.HttpHeaders.SDKSupportedCapabilities] = \
        http_constants.SDKSupportedCapabilities.PARTITION_MERGE

    # Generate a new activity ID for each request client side.
    headers[http_constants.HttpHeaders.ActivityId] = GenerateGuidId()
    if cosmos_client_connection.UseMultipleWriteLocations:
        headers[http_constants.HttpHeaders.AllowTentativeWrites] = "true"

    pre_trigger_include = options.get("preTriggerInclude")
    if pre_trigger_include:
        headers[http_constants.HttpHeaders.PreTriggerInclude] = (
            pre_trigger_include if isinstance(pre_trigger_include, str) else (",").join(pre_trigger_include)
        )

    post_trigger_include = options.get("postTriggerInclude")
    if post_trigger_include:
        headers[http_constants.HttpHeaders.PostTriggerInclude] = (
            post_trigger_include if isinstance(post_trigger_include, str) else (",").join(post_trigger_include)
        )

    if options.get("maxItemCount"):
        headers[http_constants.HttpHeaders.PageSize] = options["maxItemCount"]

    access_condition = options.get("accessCondition")
    if access_condition:
        if access_condition["type"] == "IfMatch":
            headers[http_constants.HttpHeaders.IfMatch] = access_condition["condition"]
        else:
            headers[http_constants.HttpHeaders.IfNoneMatch] = access_condition["condition"]

    if options.get("indexingDirective"):
        headers[http_constants.HttpHeaders.IndexingDirective] = options["indexingDirective"]

    # set request consistency level - if session consistency, the client should be setting this on its own
    if options.get("consistencyLevel"):
        headers[http_constants.HttpHeaders.ConsistencyLevel] = options["consistencyLevel"]

    if options.get("enableScanInQuery"):
        headers[http_constants.HttpHeaders.EnableScanInQuery] = options["enableScanInQuery"]

    if options.get("resourceTokenExpirySeconds"):
        headers[http_constants.HttpHeaders.ResourceTokenExpiry] = options["resourceTokenExpirySeconds"]

    if options.get("offerType"):
        headers[http_constants.HttpHeaders.OfferType] = options["offerType"]

    if options.get("offerThroughput"):
        headers[http_constants.HttpHeaders.OfferThroughput] = options["offerThroughput"]

    if options.get("contentType"):
        headers[http_constants.HttpHeaders.ContentType] = options['contentType']

    if options.get("isQueryPlanRequest"):
        headers[http_constants.HttpHeaders.IsQueryPlanRequest] = options['isQueryPlanRequest']

    if options.get("supportedQueryFeatures"):
        headers[http_constants.HttpHeaders.SupportedQueryFeatures] = options['supportedQueryFeatures']

    if options.get("queryVersion"):
        headers[http_constants.HttpHeaders.QueryVersion] = options['queryVersion']

    if "partitionKey" in options:
        # if partitionKey value is Undefined, serialize it as [{}] to be consistent with other SDKs.
        if isinstance(options["partitionKey"], _Undefined):
            headers[http_constants.HttpHeaders.PartitionKey] = [{}]
        # If partitionKey value is Empty, serialize it as [], which is the equivalent sent for migrated collections
        elif isinstance(options["partitionKey"], _Empty):
            headers[http_constants.HttpHeaders.PartitionKey] = []
        # else serialize using json dumps method which apart from regular values will serialize None into null
        else:
            # single partitioning uses a string and needs to be turned into a list
            is_sequence_not_string = (isinstance(options["partitionKey"], Sequence) and
                                      not isinstance(options["partitionKey"], str))

            if is_sequence_not_string and options["partitionKey"]:
                pk_val = json.dumps(list(options["partitionKey"]), separators=(',', ':'))
            else:
                pk_val = json.dumps([options["partitionKey"]])
            headers[http_constants.HttpHeaders.PartitionKey] = pk_val

    if options.get("enableCrossPartitionQuery"):
        headers[http_constants.HttpHeaders.EnableCrossPartitionQuery] = options["enableCrossPartitionQuery"]

    if options.get("populateQueryMetrics"):
        headers[http_constants.HttpHeaders.PopulateQueryMetrics] = options["populateQueryMetrics"]

    if options.get("populateIndexMetrics"):
        headers[http_constants.HttpHeaders.PopulateIndexMetrics] = options["populateIndexMetrics"]

    if options.get("responseContinuationTokenLimitInKb"):
        headers[http_constants.HttpHeaders.ResponseContinuationTokenLimitInKb] = options[
            "responseContinuationTokenLimitInKb"]

    if options.get("priorityLevel"):
        headers[http_constants.HttpHeaders.PriorityLevel] = options["priorityLevel"]

    # formatdate guarantees RFC 1123 date format regardless of current locale
    headers[http_constants.HttpHeaders.XDate] = formatdate(timeval=None, localtime=False, usegmt=True)

    if cosmos_client_connection.master_key or cosmos_client_connection.resource_tokens:
        resource_type = _internal_resourcetype(resource_type)
        authorization = _get_authorization_header(
            cosmos_client_connection, verb, path, resource_id, IsNameBased(resource_id), resource_type, headers
        )
        # urllib.quote throws when the input parameter is None
        if authorization:
            # -_.!~*'() are valid characters in url, and shouldn't be quoted.
            authorization = urllib_quote(authorization, "-_.!~*'()")
        headers[http_constants.HttpHeaders.Authorization] = authorization

    if verb in ("post", "put"):
        if not headers.get(http_constants.HttpHeaders.ContentType):
            headers[http_constants.HttpHeaders.ContentType] = _runtime_constants.MediaTypes.Json

    if not headers.get(http_constants.HttpHeaders.Accept):
        headers[http_constants.HttpHeaders.Accept] = _runtime_constants.MediaTypes.Json

    if partition_key_range_id is not None:
        headers[http_constants.HttpHeaders.PartitionKeyRangeID] = partition_key_range_id

    if client_id is not None:
        headers[http_constants.HttpHeaders.ClientId] = client_id
    elif cosmos_client_connection and cosmos_client_connection.client_id:
        headers[http_constants.HttpHeaders.ClientId] = cosmos_client_connection.client_id

    if options.get("enableScriptLogging"):
        headers[http_constants.HttpHeaders.EnableScriptLogging] = options["enableScriptLogging"]

    if options.get("offerEnableRUPerMinuteThroughput"):
        headers[http_constants.HttpHeaders.OfferIsRUPerMinuteThroughputEnabled] = options[
            "offerEnableRUPerMinuteThroughput"
        ]

    if options.get("disableRUPerMinuteUsage"):
        headers[http_constants.HttpHeaders.DisableRUPerMinuteUsage] = options["disableRUPerMinuteUsage"]

    if options.get("continuation"):
        headers[http_constants.HttpHeaders.Continuation] = options["continuation"]

    if options.get("populatePartitionKeyRangeStatistics"):
        headers[http_constants.HttpHeaders.PopulatePartitionKeyRangeStatistics] = options[
            "populatePartitionKeyRangeStatistics"
        ]

    if options.get("populateQuotaInfo"):
        headers[http_constants.HttpHeaders.PopulateQuotaInfo] = options["populateQuotaInfo"]

    if options.get("maxIntegratedCacheStaleness"):
        headers[http_constants.HttpHeaders.DedicatedGatewayCacheStaleness] = options["maxIntegratedCacheStaleness"]

    if options.get("autoUpgradePolicy"):
        headers[http_constants.HttpHeaders.AutoscaleSettings] = options["autoUpgradePolicy"]

    if options.get("correlatedActivityId"):
        headers[http_constants.HttpHeaders.CorrelatedActivityId] = options["correlatedActivityId"]

    if options.get("throughputBucket"):
        headers[http_constants.HttpHeaders.ThroughputBucket] = options["throughputBucket"]

    if resource_type == "docs" and verb != "get":
        if "responsePayloadOnWriteDisabled" in options:
            responsePayloadOnWriteDisabled = options["responsePayloadOnWriteDisabled"]
        else:
            responsePayloadOnWriteDisabled = cosmos_client_connection.connection_policy.ResponsePayloadOnWriteDisabled

        if responsePayloadOnWriteDisabled:
            headers[http_constants.HttpHeaders.Prefer] = "return=minimal"

    # If it is an operation at the container level, verify the rid of the container to see if the cache needs to be
    # refreshed.
    if resource_type != 'dbs' and options.get("containerRID"):
        headers[http_constants.HttpHeaders.IntendedCollectionRID] = options["containerRID"]

    if resource_type == "":
        resource_type = http_constants.ResourceType.DatabaseAccount
    headers[http_constants.HttpHeaders.ThinClientProxyResourceType] = resource_type
    headers[http_constants.HttpHeaders.ThinClientProxyOperationType] = operation_type

    return headers

def _is_session_token_request(
        cosmos_client_connection: Union["CosmosClientConnection", "AsyncClientConnection"],
        headers: dict,
        request_object: "RequestObject") -> bool:
    consistency_level = headers.get(http_constants.HttpHeaders.ConsistencyLevel)
    # Figure out if consistency level for this request is session
    is_session_consistency = consistency_level == documents.ConsistencyLevel.Session

    # Verify that it is not a metadata request, and that it is either a read request, batch request, or an account
    # configured to use multiple write regions. Batch requests are special-cased because they can contain both read and
    # write operations, and we want to use session consistency for the read operations.
    return (is_session_consistency is True and not IsMasterResource(request_object.resource_type)
            and (documents._OperationType.IsReadOnlyOperation(request_object.operation_type)
                 or request_object.operation_type == "Batch"
                 or cosmos_client_connection._global_endpoint_manager.can_use_multiple_write_locations(request_object)))


def set_session_token_header(
        cosmos_client_connection: Union["CosmosClientConnection", "AsyncClientConnection"],
        headers: dict,
        path: str,
        request_object: "RequestObject",
        options: Mapping[str, Any],
        partition_key_range_id: Optional[str] = None) -> None:
    # set session token if required
    if _is_session_token_request(cosmos_client_connection, headers, request_object):
        # if there is a token set via option, then use it to override default
        if options.get("sessionToken"):
            headers[http_constants.HttpHeaders.SessionToken] = options["sessionToken"]
        else:
            # check if the client's default consistency is session (and request consistency level is same),
            # then update from session container
            if headers[http_constants.HttpHeaders.ConsistencyLevel] == documents.ConsistencyLevel.Session and \
                    cosmos_client_connection.session:
                # urllib_unquote is used to decode the path, as it may contain encoded characters
                path = urllib_unquote(path)
                # populate session token from the client's session container
                session_token = (
                    cosmos_client_connection.session.get_session_token(path,
                                                                options.get('partitionKey'),
                                                                cosmos_client_connection._container_properties_cache,
                                                                cosmos_client_connection._routing_map_provider,
                                                                partition_key_range_id,
                                                                options))
                if session_token != "":
                    headers[http_constants.HttpHeaders.SessionToken] = session_token

async def set_session_token_header_async(
        cosmos_client_connection: Union["CosmosClientConnection", "AsyncClientConnection"],
        headers: dict,
        path: str,
        request_object: "RequestObject",
        options: Mapping[str, Any],
        partition_key_range_id: Optional[str] = None) -> None:
    # set session token if required
    if _is_session_token_request(cosmos_client_connection, headers, request_object):
        # if there is a token set via option, then use it to override default
        if options.get("sessionToken"):
            headers[http_constants.HttpHeaders.SessionToken] = options["sessionToken"]
        else:
            # check if the client's default consistency is session (and request consistency level is same),
            # then update from session container
            if headers[http_constants.HttpHeaders.ConsistencyLevel] == documents.ConsistencyLevel.Session and \
                    cosmos_client_connection.session:
                # populate session token from the client's session container
                # urllib_unquote is used to decode the path, as it may contain encoded characters
                path = urllib_unquote(path)
                session_token = \
                    await cosmos_client_connection.session.get_session_token_async(path,
                                                                options.get('partitionKey'),
                                                                cosmos_client_connection._container_properties_cache,
                                                                cosmos_client_connection._routing_map_provider,
                                                                partition_key_range_id,
                                                                options)
                if session_token != "":
                    headers[http_constants.HttpHeaders.SessionToken] = session_token

def GetResourceIdOrFullNameFromLink(resource_link: str) -> str:
    """Gets resource id or full name from resource link.

    :param str resource_link:
    :return: The resource id or full name from the resource link.
    :rtype: str
    """
    # For named based, the resource link is the full name
    if IsNameBased(resource_link):
        return TrimBeginningAndEndingSlashes(resource_link)

    # Padding the resource link with leading and trailing slashes if not already
    if resource_link[-1] != "/":
        resource_link = resource_link + "/"

    if resource_link[0] != "/":
        resource_link = "/" + resource_link

    # The path will be in the form of
    # /[resourceType]/[resourceId]/ .... /[resourceType]/[resourceId]/ or
    # /[resourceType]/[resourceId]/ .... /[resourceType]/
    # The result of split will be in the form of
    # ["", [resourceType], [resourceId] ... ,[resourceType], [resourceId], ""]
    # In the first case, to extract the resourceId it will the element
    # before last ( at length -2 ) and the the type will before it
    # ( at length -3 )
    # In the second case, to extract the resource type it will the element
    # before last ( at length -2 )
    path_parts = resource_link.split("/")
    if len(path_parts) % 2 == 0:
        # request in form
        # /[resourceType]/[resourceId]/ .... /[resourceType]/[resourceId]/.
        return str(path_parts[-2])
    raise ValueError("Failed Parsing ResourceID from link: {0}".format(resource_link))


def GenerateGuidId() -> str:
    """Gets a random GUID.

    Note that here we use python's UUID generation library. Basically UUID
    is the same as GUID when represented as a string.

    :return:
        The generated random GUID.
    :rtype: str
    """
    return str(uuid.uuid4())


def GetPathFromLink(resource_link: str, resource_type: str = "") -> str:
    """Gets path from resource link with optional resource type

    :param str resource_link:
    :param str resource_type:
    :return: Path from resource link with resource type appended (if provided).
    :rtype: str
    """
    resource_link = TrimBeginningAndEndingSlashes(resource_link)

    if IsNameBased(resource_link):
        # Replace special characters in string using the %xx escape. For example,
        # space(' ') would be replaced by %20 This function is intended for quoting
        # the path section of the URL and excludes '/' to be quoted as that's the
        # default safe char
        resource_link = urllib_quote(resource_link)

    # Padding leading and trailing slashes to the path returned both for name based and resource id based links
    if resource_type:
        return "/" + resource_link + "/" + resource_type + "/"
    return "/" + resource_link + "/"


def IsNameBased(link: Optional[str]) -> bool:
    """Finds whether the link is name based or not

    :param str link:

    :return:
        True if link is name-based; otherwise, False.
    :rtype: boolean
    """
    if not link:
        return False

    # trimming the leading "/"
    if link.startswith("/") and len(link) > 1:
        link = link[1:]

    # Splitting the link(separated by "/") into parts
    parts = link.split("/")

    # First part should be "dbs"
    if not (parts and parts[0].lower() == "dbs"):
        return False

    # The second part is the database id(ResourceID or Name) and cannot be empty
    if len(parts) < 2 or not parts[1]:
        return False

    # Either ResourceID or database name
    databaseID = parts[1]

    # Length of databaseID(in case of ResourceID) is always 8
    if len(databaseID) != 8:
        return True

    return not IsValidBase64String(str(databaseID))


def IsMasterResource(resourceType: str) -> bool:
    return resourceType in (
        http_constants.ResourceType.Offer,
        http_constants.ResourceType.Database,
        http_constants.ResourceType.User,
        http_constants.ResourceType.Permission,
        http_constants.ResourceType.Topology,
        http_constants.ResourceType.DatabaseAccount,
        http_constants.ResourceType.PartitionKeyRange,
        http_constants.ResourceType.Collection,
    )


def IsDatabaseLink(link: str) -> bool:
    """Finds whether the link is a database Self Link or a database ID based link

    :param str link: Link to analyze
    :return: True or False.
    :rtype: boolean
    """
    if not link:
        return False

    # trimming the leading and trailing "/" from the input string
    link = TrimBeginningAndEndingSlashes(link)

    # Splitting the link(separated by "/") into parts
    parts = link.split("/")

    if len(parts) != 2:
        return False

    # First part should be "dbs"
    if not parts[0] or not parts[0].lower() == "dbs":
        return False

    # The second part is the database id(ResourceID or Name) and cannot be empty
    if not parts[1]:
        return False

    return True


def IsItemContainerLink(link: str) -> bool:  # pylint: disable=too-many-return-statements
    """Finds whether the link is a document colllection Self Link or a document colllection ID based link

    :param str link: Link to analyze
    :return: True or False.
    :rtype: boolean
    """
    if not link:
        return False

    # trimming the leading and trailing "/" from the input string
    link = TrimBeginningAndEndingSlashes(link)

    # Splitting the link(separated by "/") into parts
    parts = link.split("/")

    if len(parts) != 4:
        return False

    # First part should be "dbs"
    if not parts[0] or not parts[0].lower() == "dbs":
        return False

    # Third part should be "colls"
    if not parts[2] or not parts[2].lower() == "colls":
        return False

    # The second part is the database id(ResourceID or Name) and cannot be empty
    if not parts[1]:
        return False

    # The fourth part is the document collection id(ResourceID or Name) and cannot be empty
    if not parts[3]:
        return False

    return True


def GetItemContainerInfo(self_link: str, alt_content_path: str, resource_id: str) -> Tuple[str, str]:
    """Given the self link and alt_content_path from the response header and
    result extract the collection name and collection id.

    Every response header has an alt-content-path that is the owner's path in
    ASCII. For document create / update requests, this can be used to get the
    collection name, but for collection create response, we can't use it.

    :param str self_link:
        Self link of the resource, as obtained from response result.
    :param str alt_content_path:
        Owner path of the resource, as obtained from response header.
    :param str resource_id:
        'id' as returned from the response result. This is only used if it is
        deduced that the request was to create a collection.
    :return: tuple of (collection rid, collection name)
    :rtype: tuple
    """

    self_link = TrimBeginningAndEndingSlashes(self_link) + "/"

    end_index = IndexOfNth(self_link, "/", 4)
    start_index = IndexOfNth(self_link, "/", 3)

    if start_index != -1 and end_index != -1:
        # parse only the collection rid from the path as it's unique across databases
        collection_rid = self_link[start_index + 1:end_index]

        if "colls" in self_link:
            # this is a collection request
            index_second_slash = IndexOfNth(alt_content_path, "/", 2)
            if index_second_slash == -1:
                collection_name = alt_content_path + "/colls/" + urllib_quote(resource_id)
                return collection_rid, collection_name
            collection_name = alt_content_path
            return collection_rid, collection_name
        raise ValueError(
            "Response Not from Server Partition, self_link: {0}, alt_content_path: {1}, id: {2}".format(
                self_link, alt_content_path, resource_id
            )
        )

    raise ValueError("Unable to parse document collection link from " + self_link)


def GetItemContainerLink(link: str) -> str:
    """Gets the document collection link.

    :param str link: Resource link
    :return: Document collection link.
    :rtype: str
    """
    link = TrimBeginningAndEndingSlashes(link) + "/"

    index = IndexOfNth(link, "/", 4)

    if index != -1:
        return link[0:index]
    raise ValueError("Unable to parse document collection link from " + link)


def IndexOfNth(s: str, value: str, n: int) -> int:
    """Gets the index of Nth occurrence of a given character in a string.

    :param str s: Input string
    :param str value: Input char to be searched.
    :param int n: Nth occurrence of char to be searched.
    :return: Index of the Nth occurrence in the string.
    :rtype: int
    """
    remaining = n
    for i, elt in enumerate(s):
        if elt == value:
            remaining -= 1
            if remaining == 0:
                return i
    return -1


def IsValidBase64String(string_to_validate: str) -> bool:
    """Verifies if a string is a valid Base64 encoded string, after
    replacing '-' with '/'

    :param string string_to_validate: String to validate.
    :return: Whether given string is a valid base64 string or not.
    :rtype: str
    """
    # '-' is not supported char for decoding in Python(same as C# and Java) which has
    # similar logic while parsing ResourceID generated by backend
    try:
        buffer = base64.standard_b64decode(string_to_validate.replace("-", "/"))
        if len(buffer) != 4:
            return False
    except Exception as e:  # pylint: disable=broad-except
        if isinstance(e, binascii.Error):
            return False
        raise e
    return True


def TrimBeginningAndEndingSlashes(path: str) -> str:
    """Trims beginning and ending slashes

    :param str path:

    :return:
        Path with beginning and ending slashes trimmed.
    :rtype: str
    """
    if path.startswith("/"):
        # Returns substring starting from index 1 to end of the string
        path = path[1:]

    if path.endswith("/"):
        # Returns substring starting from beginning to last but one char in the string
        path = path[:-1]

    return path


# Parses the paths into a list of token each representing a property
def ParsePaths(paths: list[str]) -> list[str]:
    segmentSeparator = "/"
    tokens = []
    for path in paths:
        currentIndex = 0

        while currentIndex < len(path):
            if path[currentIndex] != segmentSeparator:
                raise ValueError(f"Invalid path character at index {currentIndex}")

            currentIndex += 1
            if currentIndex == len(path):
                break

            # " and ' are treated specially in the sense that they can have the / (segment separator)
            # between them which is considered part of the token
            if path[currentIndex] == '"' or path[currentIndex] == "'":
                quote = path[currentIndex]
                newIndex = currentIndex + 1

                while True:
                    newIndex = path.find(quote, newIndex)
                    if newIndex == -1:
                        raise ValueError(f"Invalid path character at index {currentIndex}")

                    # check if the quote itself is escaped by a preceding \ in which case it's part of the token
                    if path[newIndex - 1] != "\\":
                        break
                    newIndex += 1

                # This will extract the token excluding the quote chars
                token = path[currentIndex + 1: newIndex]
                tokens.append(token)
                currentIndex = newIndex + 1
            else:
                newIndex = path.find(segmentSeparator, currentIndex)
                token = None
                if newIndex == -1:
                    # This will extract the token from currentIndex to end of the string
                    token = path[currentIndex:]
                    currentIndex = len(path)
                else:
                    # This will extract the token from currentIndex to the char before the segmentSeparator
                    token = path[currentIndex:newIndex]
                    currentIndex = newIndex

                token = token.strip()
                tokens.append(token)

    return tokens


def create_scope_from_url(url: str) -> str:
    parsed_url = urlsplit(url)
    if not parsed_url.scheme or not parsed_url.hostname:
        raise ValueError(f"Invalid URL scheme or hostname: {parsed_url!r}")
    return parsed_url.scheme + "://" + parsed_url.hostname + "/.default"


def validate_cache_staleness_value(max_integrated_cache_staleness: Any) -> None:
    int(max_integrated_cache_staleness)  # Will throw error if data type cant be converted to int
    if max_integrated_cache_staleness < 0:
        raise ValueError("Parameter 'max_integrated_cache_staleness_in_ms' can only be an "
                         "integer greater than or equal to zero")


def _validate_resource(resource: Mapping[str, Any]) -> None:
    id_: Optional[str] = resource.get("id")
    if id_:
        try:
            if _VALID_COSMOS_RESOURCE.match(id_) is None:
                raise ValueError("Id contains illegal chars.")
            if id_[-1] in [" ", "\n"]:
                raise ValueError("Id ends with a space or newline.")
        except TypeError as e:
            raise TypeError("Id type must be a string.") from e


def _stringify_auto_scale(offer: ThroughputProperties) -> str:
    auto_scale_params: Optional[dict[str, Union[None, int, dict[str, Any]]]] = None
    max_throughput = offer.auto_scale_max_throughput
    increment_percent = offer.auto_scale_increment_percent
    auto_scale_params = {"maxThroughput": max_throughput}
    if increment_percent is not None:
        auto_scale_params["autoUpgradePolicy"] = {"throughputPolicy": {"incrementPercent": increment_percent}}
    auto_scale_settings = json.dumps(auto_scale_params)
    return auto_scale_settings


def _set_throughput_options(offer: Optional[Union[int, ThroughputProperties]], request_options: dict[str, Any]) -> None:
    if isinstance(offer, int):
        request_options["offerThroughput"] = offer
    elif offer is not None:
        try:
            max_throughput = offer.auto_scale_max_throughput
            increment_percent = offer.auto_scale_increment_percent

            if max_throughput is not None:
                request_options['autoUpgradePolicy'] = _stringify_auto_scale(offer=offer)
            elif increment_percent:
                raise ValueError("auto_scale_max_throughput must be supplied in "
                                 "conjunction with auto_scale_increment_percent")
            if offer.offer_throughput:
                request_options["offerThroughput"] = offer.offer_throughput
        except AttributeError as e:
            raise TypeError("offer_throughput must be int or an instance of ThroughputProperties") from e


def _deserialize_throughput(throughput: list[dict[str, dict[str, Any]]]) -> ThroughputProperties:
    properties = throughput[0]
    offer_autopilot: Optional[dict[str, Any]] = properties['content'].get('offerAutopilotSettings')
    if offer_autopilot and 'autoUpgradePolicy' in offer_autopilot:
        return ThroughputProperties(
            properties=properties,
            auto_scale_max_throughput=offer_autopilot['maxThroughput'],
            auto_scale_increment_percent=offer_autopilot['autoUpgradePolicy']['throughputPolicy']['incrementPercent']
        )
    if offer_autopilot:
        return ThroughputProperties(
            properties=properties,
            auto_scale_max_throughput=offer_autopilot['maxThroughput']
        )
    return ThroughputProperties(
        offer_throughput=properties["content"]["offerThroughput"],
        properties=properties
    )


def _replace_throughput(
    throughput: Union[int, ThroughputProperties],
    new_throughput_properties: dict[str, Any]
) -> None:
    if isinstance(throughput, int):
        new_throughput_properties["content"]["offerThroughput"] = throughput
    else:
        try:
            max_throughput = throughput.auto_scale_max_throughput
            increment_percent = throughput.auto_scale_increment_percent
            if max_throughput is not None:
                new_throughput_properties['content']['offerAutopilotSettings']['maxThroughput'] = max_throughput
                if increment_percent:
                    new_throughput_properties['content']['offerAutopilotSettings']['autoUpgradePolicy']['throughputPolicy']['incrementPercent'] = increment_percent  # pylint: disable=line-too-long
            if throughput.offer_throughput:
                new_throughput_properties["content"]["offerThroughput"] = throughput.offer_throughput
        except AttributeError as e:
            raise TypeError("offer_throughput must be int or an instance of ThroughputProperties") from e


def _internal_resourcetype(resource_type: str) -> str:
    """Partition key is used as the resource type for deleting all items by partition key in other SDKs,
    but the colls (collection) resource type needs to be sent for the feature to work. In order to keep it consistent
    with other SDKs, we switch it here.
    :param str resource_type: the resource type
    :return: the resource type after checking if we're doing partition key delete.
    :rtype: str
    """
    if resource_type.lower() == "partitionkey":
        return "colls"
    return resource_type


def _populate_batch_headers(current_headers: dict[str, Any]) -> None:
    current_headers[http_constants.HttpHeaders.IsBatchRequest] = True
    current_headers[http_constants.HttpHeaders.IsBatchAtomic] = True
    current_headers[http_constants.HttpHeaders.ShouldBatchContinueOnError] = False


def _format_batch_operations(
    operations: Sequence[Union[Tuple[str, Tuple[Any, ...]], Tuple[str, Tuple[Any, ...], dict[str, Any]]]]
) -> list[dict[str, Any]]:
    final_operations = []
    for index, batch_operation in enumerate(operations):
        try:
            operation_type = batch_operation[0]
            args = batch_operation[1]
        except IndexError as e:
            raise IndexError(f"Operation {index} in batch is missing a field.") from e
        try:
            kwargs = batch_operation[2]  # type: ignore[misc]
        except IndexError:
            kwargs = {}

        if len(args) == 1:
            if operation_type.lower() == "create":
                operation = {"operationType": "Create",
                             "resourceBody": args[0]}
            elif operation_type.lower() == "upsert":
                operation = {"operationType": "Upsert",
                             "resourceBody": args[0]}
            elif operation_type.lower() == "read":
                operation = {"operationType": "Read",
                             "id": args[0]}
            elif operation_type.lower() == "delete":
                operation = {"operationType": "Delete",
                             "id": args[0]}
        elif len(args) == 2:
            if operation_type.lower() == "replace":
                operation = {"operationType": "Replace",
                             "id": args[0],
                             "resourceBody": args[1]}
            elif operation_type.lower() == "patch":
                operation = {"operationType": "Patch",
                             "id": args[0],
                             "resourceBody": {"operations": args[1]}}
                filter_predicate = kwargs.pop("filter_predicate", None)
                if filter_predicate is not None:
                    operation["resourceBody"]["condition"] = filter_predicate
        else:
            raise AttributeError(
                f"Operation type or args passed in not recognized for operation with index {index}."
            )

        if_match_etag = kwargs.pop("if_match_etag", None)
        if_none_match_etag = kwargs.pop("if_none_match_etag", None)
        if if_match_etag is not None:
            operation["ifMatch"] = if_match_etag
        elif if_none_match_etag is not None:
            operation["ifNoneMatch"] = if_none_match_etag

        final_operations.append(operation)

    return final_operations


def _build_properties_cache(properties: dict[str, Any], container_link: str) -> dict[str, Any]:
    return {
        "_self": properties.get("_self", None), "_rid": properties.get("_rid", None),
        "partitionKey": properties.get("partitionKey", None), "container_link": container_link
    }

def format_pk_range_options(query_options: Mapping[str, Any]) -> dict[str, Any]:
    """Formats the partition key range options to be used internally from the query ones.
    :param dict query_options: The query options being used.
    :return: The relevant partition key range options.
    :rtype: dict
    """
    pk_range_options: dict[str, Any] = {}
    if query_options is not None:
        if "containerRID" in query_options:
            pk_range_options["containerRID"] = query_options["containerRID"]
        if "excludedLocations" in query_options:
            pk_range_options["excludedLocations"] = query_options["excludedLocations"]
    return pk_range_options



# The MIT License (MIT)
# Copyright (c) 2014 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Internal class for collection routing map implementation in the Azure Cosmos
database service.
"""

import bisect
from azure.cosmos._routing import routing_range
from azure.cosmos._routing.routing_range import PartitionKeyRange


class CollectionRoutingMap(object):
    """Stores partition key ranges in an efficient way with some additional
    information and provides convenience methods for working with set of ranges.
    """

    MinimumInclusiveEffectivePartitionKey = ""
    MaximumExclusiveEffectivePartitionKey = "FF"

    def __init__(
        self, range_by_id, range_by_info, ordered_partition_key_ranges, ordered_partition_info, collection_unique_id
    ):
        self._rangeById = range_by_id
        self._rangeByInfo = range_by_info
        self._orderedPartitionKeyRanges = ordered_partition_key_ranges

        self._orderedRanges = [
            routing_range.Range(pkr[PartitionKeyRange.MinInclusive], pkr[PartitionKeyRange.MaxExclusive], True, False)
            for pkr in ordered_partition_key_ranges
        ]
        self._orderedPartitionInfo = ordered_partition_info
        self._collectionUniqueId = collection_unique_id

    @classmethod
    def CompleteRoutingMap(cls, partition_key_range_info_tuple_list, collection_unique_id):
        rangeById = {}
        rangeByInfo = {}

        sortedRanges = []
        for r in partition_key_range_info_tuple_list:
            rangeById[r[0][PartitionKeyRange.Id]] = r
            rangeByInfo[r[1]] = r[0]
            sortedRanges.append(r)

        sortedRanges.sort(key=lambda r: r[0][PartitionKeyRange.MinInclusive])
        partitionKeyOrderedRange = [r[0] for r in sortedRanges]
        orderedPartitionInfo = [r[1] for r in sortedRanges]

        if not CollectionRoutingMap.is_complete_set_of_range(partitionKeyOrderedRange):
            return None
        return cls(rangeById, rangeByInfo, partitionKeyOrderedRange, orderedPartitionInfo, collection_unique_id)

    def get_ordered_partition_key_ranges(self):
        """Gets the ordered partition key ranges

        :return: Ordered list of partition key ranges.
        :rtype: list
        """
        return self._orderedPartitionKeyRanges

    def get_range_by_effective_partition_key(self, effective_partition_key_value):
        """Gets the range containing the given partition key

        :param str effective_partition_key_value: The partition key value.
        :return: The partition key range.
        :rtype: dict
        """
        if CollectionRoutingMap.MinimumInclusiveEffectivePartitionKey == effective_partition_key_value:
            return self._orderedPartitionKeyRanges[0]

        if CollectionRoutingMap.MaximumExclusiveEffectivePartitionKey == effective_partition_key_value:
            return None

        sortedLow = [(r.min, not r.isMinInclusive) for r in self._orderedRanges]

        index = bisect.bisect_right(sortedLow, (effective_partition_key_value, True))
        if index > 0:
            index = index - 1
        return self._orderedPartitionKeyRanges[index]

    def get_range_by_partition_key_range_id(self, partition_key_range_id):
        """Gets the partition key range given the partition key range id

        :param str partition_key_range_id: The partition key range id.
        :return: The partition key range.
        :rtype: dict
        """
        t = self._rangeById.get(partition_key_range_id)

        if t is None:
            return None
        return t[0]

    def get_overlapping_ranges(self, provided_partition_key_ranges):
        """Gets the partition key ranges overlapping the provided ranges

        :param list provided_partition_key_ranges: List of partition key ranges.
        :return: List of partition key ranges, where each is a dict.
        :rtype: list
        """

        if isinstance(provided_partition_key_ranges, routing_range.Range):
            return self.get_overlapping_ranges([provided_partition_key_ranges])

        minToPartitionRange = {}

        sortedLow = [(r.min, not r.isMinInclusive) for r in self._orderedRanges]
        sortedHigh = [(r.max, r.isMaxInclusive) for r in self._orderedRanges]

        for providedRange in provided_partition_key_ranges:
            minIndex = bisect.bisect_right(sortedLow, (providedRange.min, not providedRange.isMinInclusive))
            if minIndex > 0:
                minIndex = minIndex - 1

            maxIndex = bisect.bisect_left(sortedHigh, (providedRange.max, providedRange.isMaxInclusive))
            if maxIndex >= len(sortedHigh):
                maxIndex = maxIndex - 1

            for i in range(minIndex, maxIndex + 1):
                if routing_range.Range.overlaps(self._orderedRanges[i], providedRange):
                    minToPartitionRange[
                        self._orderedPartitionKeyRanges[i][PartitionKeyRange.MinInclusive]
                    ] = self._orderedPartitionKeyRanges[i]

        overlapping_partition_key_ranges = list(minToPartitionRange.values())

        def getKey(r):
            return r[PartitionKeyRange.MinInclusive]

        overlapping_partition_key_ranges.sort(key=getKey)
        return overlapping_partition_key_ranges

    @staticmethod
    def is_complete_set_of_range(ordered_partition_key_range_list):
        isComplete = False
        if ordered_partition_key_range_list:

            firstRange = ordered_partition_key_range_list[0]
            lastRange = ordered_partition_key_range_list[-1]
            isComplete = (
                firstRange[PartitionKeyRange.MinInclusive] == CollectionRoutingMap.MinimumInclusiveEffectivePartitionKey
            )
            isComplete &= (
                lastRange[PartitionKeyRange.MaxExclusive] == CollectionRoutingMap.MaximumExclusiveEffectivePartitionKey
            )

            for i in range(1, len(ordered_partition_key_range_list)):
                previousRange = ordered_partition_key_range_list[i - 1]
                currentRange = ordered_partition_key_range_list[i]
                isComplete &= (
                    previousRange[PartitionKeyRange.MaxExclusive] == currentRange[PartitionKeyRange.MinInclusive]
                )

                if not isComplete:
                    if previousRange[PartitionKeyRange.MaxExclusive] > currentRange[PartitionKeyRange.MinInclusive]:
                        raise ValueError("Ranges overlap")
                    break

        return isComplete



# The MIT License (MIT)
# Copyright (c) 2014 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Internal class for client side partition implementation in the Azure Cosmos
database service.
"""


class Partition(object):
    """A class that holds the hash value and node name for a partition.
    """

    def __init__(self, hash_value=None, node=None):
        self.hash_value = hash_value
        self.node = node

    def GetNode(self):
        """Gets the name of the node(collection) for this object.
        """
        return self.node

    def __eq__(self, other):
        return (self.hash_value == other.hash_value) and (self.node == other.node)

    def __lt__(self, other):
        if self == other:
            return False

        return self.CompareTo(other.hash_value) < 0

    def CompareTo(self, other_hash_value):
        """Compare the passed hash value with the hash value of this object.
        :param List[int] other_hash_value: the hash value to be compared
        :returns: an integer stating the result of the comparison
         -1 if other_hash_value is greater than self.hash_value, 1 if other_hash_value is less than self.hash_value,
          0 if they're both equal
        :rtype: int
        """
        if len(self.hash_value) != len(other_hash_value):
            raise ValueError("Length of hashes doesn't match.")

        # The hash byte array that is returned from ComputeHash method has the MSB at the end of the array
        # so comparing the bytes from the end for compare operations.
        for i in range(0, len(self.hash_value)):
            if self.hash_value[len(self.hash_value) - i - 1] < other_hash_value[len(self.hash_value) - i - 1]:
                return -1
            if self.hash_value[len(self.hash_value) - i - 1] > other_hash_value[len(self.hash_value) - i - 1]:
                return 1
        return 0



# The MIT License (MIT)
# Copyright (c) 2021 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Create throughput properties in the Azure Cosmos DB SQL API service.
"""


from typing import Optional, Any, Mapping, TYPE_CHECKING

if TYPE_CHECKING:
    from ._cosmos_responses import CosmosDict


class ThroughputProperties:
    """Represents the throughput properties in an Azure Cosmos DB SQL API container.

    To read and update throughput properties, use the associated methods on the :class:`Container`.
    If configuring auto-scale, `auto_scale_max_throughput` needs to be set and
    `auto_scale_increment_percent` can also be set in conjunction with it.
    The value of `offer_throughput` will not be allowed to be set in conjunction with the auto-scale settings.

    :keyword int offer_throughput: The provisioned throughput in request units per second as a number.
    :keyword int auto_scale_max_throughput: The max auto-scale throughput. It should have a valid throughput
     value between 1000 and 1000000 inclusive, in increments of 1000.
    :keyword int auto_scale_increment_percent: is the % from the base selected RU it increases at a given time,
     the increment percent should be greater than or equal to zero.
    """

    def __init__(self, *args, **kwargs) -> None:
        self.offer_throughput: Optional[int] = args[0] if args else kwargs.get('offer_throughput')
        self.properties: Optional["CosmosDict"] = args[1] if len(args) > 1 else kwargs.get('properties')
        self.auto_scale_max_throughput: Optional[int] = kwargs.get('auto_scale_max_throughput')
        self.auto_scale_increment_percent: Optional[int] = kwargs.get('auto_scale_increment_percent')

    def get_response_headers(self) -> Mapping[str, Any]:
        """Returns a copy of the response headers associated to this response

        :return: Dict of response headers
        :rtype: ~azure.core.utils.CaseInsensitiveDict
        """
        if self.properties is None:
            return {}
        try:
            return self.properties.get_response_headers()
        except AttributeError:
            return {}

Offer = ThroughputProperties



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""
FILE: autoscale_throughput_management_async.py

DESCRIPTION:
    This async sample demonstrates how to manage autoscale throughput settings for
    Azure Cosmos DB databases and containers. Autoscale allows you to automatically
    scale throughput based on usage, providing cost optimization and performance flexibility.

    Key concepts covered:
    - Creating databases and containers with autoscale throughput (async)
    - Reading autoscale throughput settings (async)
    - Updating autoscale maximum throughput (async)
    - Understanding autoscale increment percentage

USAGE:
    python autoscale_throughput_management_async.py

    Set the environment variables with your own values before running:
    1) ACCOUNT_HOST - the Cosmos DB account endpoint
    2) ACCOUNT_KEY - the Cosmos DB account primary key
"""

import asyncio
from azure.cosmos.aio import CosmosClient
import azure.cosmos.exceptions as exceptions
from azure.cosmos.partition_key import PartitionKey
from azure.cosmos import ThroughputProperties

import config

# ----------------------------------------------------------------------------------------------------------
# Prerequisites -
#
# 1. An Azure Cosmos account -
#    https://docs.microsoft.com/azure/cosmos-db/create-sql-api-python#create-a-database-account
#
# 2. Microsoft Azure Cosmos PyPi package -
#    https://pypi.python.org/pypi/azure-cosmos/
# ----------------------------------------------------------------------------------------------------------
# Sample - demonstrates async autoscale throughput management for databases and containers
# ----------------------------------------------------------------------------------------------------------
# Note -
#
# Running this sample will create (and delete) multiple Containers on your account.
# Each time a Container is created the account will be billed for 1 hour of usage based on
# the provisioned throughput (RU/s) of that account.
# ----------------------------------------------------------------------------------------------------------

HOST = config.settings['host']
MASTER_KEY = config.settings['master_key']
DATABASE_ID = config.settings['database_id']
CONTAINER_ID = config.settings['container_id']


async def create_database_with_autoscale(client, database_id):
    """
    Create a database with autoscale throughput asynchronously.
    Setting throughput settings, like autoscale, on a database level is *not* recommended,
    and should only be done if you are aware of the implications of shared throughput across containers.

    Autoscale throughput automatically scales between 10% and 100% of the maximum throughput
    based on your workload demands.
    
    Args:
        client: CosmosClient instance
        database_id: ID for the database
    """
    print("\nCreate Database with Autoscale Throughput (Async)")
    print("=" * 70)
    
    try:
        # Create database with autoscale - max throughput of 4000 RU/s
        # The database will scale between 400 RU/s (10%) and 4000 RU/s (100%)
        database = await client.create_database(
            id=database_id,
            offer_throughput=ThroughputProperties(
                auto_scale_max_throughput=4000,
                auto_scale_increment_percent=0
            )
        )
        
        print(f"Database '{database_id}' created with autoscale")
        print(f"  - Maximum throughput: 4000 RU/s")
        print(f"  - Minimum throughput: 400 RU/s (10% of max)")
        print(f"  - Auto-scales based on usage between min and max")
        
        return database
        
    except exceptions.CosmosResourceExistsError:
        print(f"Database '{database_id}' already exists")
        return client.get_database_client(database_id)


async def create_container_with_autoscale(database, container_id):
    """
    Create a container with autoscale throughput asynchronously.
    
    Container-level autoscale provides dedicated throughput for a specific container,
    independent of the database throughput.
    
    Args:
        database: DatabaseProxy instance
        container_id: ID for the container
    """
    print("\nCreate Container with Autoscale Throughput (Async)")
    print("=" * 70)
    
    try:
        # Create container with autoscale - max throughput of 5000 RU/s
        # auto_scale_increment_percent=0 means default scaling behavior
        container = await database.create_container(
            id=container_id,
            partition_key=PartitionKey(path='/id'),
            offer_throughput=ThroughputProperties(
                auto_scale_max_throughput=5000,
                auto_scale_increment_percent=0
            )
        )
        
        print(f"Container '{container_id}' created with autoscale")
        print(f"  - Maximum throughput: 5000 RU/s")
        print(f"  - Minimum throughput: 500 RU/s (10% of max)")
        print(f"  - Scales automatically based on workload")
        
        return container
        
    except exceptions.CosmosResourceExistsError:
        print(f"Container '{container_id}' already exists")
        return database.get_container_client(container_id)


async def read_autoscale_throughput(database, container):
    """
    Read and display autoscale throughput settings for database and container asynchronously.
    
    The throughput properties reveal:
    - Whether autoscale is enabled
    - Maximum throughput setting
    - Current throughput (if available)
    
    Args:
        database: DatabaseProxy instance
        container: ContainerProxy instance
    """
    print("\nRead Autoscale Throughput Settings (Async)")
    print("=" * 70)
    
    try:
        # Read database throughput
        db_offer = await database.get_throughput()
        print(f"\nDatabase '{database.id}' throughput:")
        
        autopilot_settings = db_offer.properties.get('content', {}).get('offerAutopilotSettings')
        if autopilot_settings:
            max_throughput = autopilot_settings.get('maxThroughput')
            print(f"  - Autoscale enabled: Yes")
            print(f"  - Maximum throughput: {max_throughput} RU/s")
            print(f"  - Minimum throughput: {max_throughput // 10} RU/s")
            print(f"  - Increment percent: {autopilot_settings.get('autoUpgradePolicy', {}).get('throughputPolicy', {}).get('incrementPercent', 0)}")
        else:
            throughput = db_offer.properties.get('content', {}).get('offerThroughput')
            print(f"  - Autoscale enabled: No")
            print(f"  - Manual throughput: {throughput} RU/s")
            
    except exceptions.CosmosHttpResponseError as e:
        print(f"Database throughput error: {e.message}")
    
    try:
        # Read container throughput
        container_offer = await container.get_throughput()
        print(f"\nContainer '{container.id}' throughput:")
        
        autopilot_settings = container_offer.properties.get('content', {}).get('offerAutopilotSettings')
        if autopilot_settings:
            max_throughput = autopilot_settings.get('maxThroughput')
            print(f"  - Autoscale enabled: Yes")
            print(f"  - Maximum throughput: {max_throughput} RU/s")
            print(f"  - Minimum throughput: {max_throughput // 10} RU/s")
        else:
            throughput = container_offer.properties.get('content', {}).get('offerThroughput')
            print(f"  - Autoscale enabled: No")
            print(f"  - Manual throughput: {throughput} RU/s")
            
    except exceptions.CosmosHttpResponseError as e:
        print(f"Container throughput error: {e.message}")


async def update_autoscale_max_throughput(container, new_max_throughput):
    """
    Update the maximum throughput for an autoscale-enabled container asynchronously.
    
    This changes the upper limit of the autoscale range. The minimum throughput
    will automatically adjust to 10% of the new maximum.
    
    Args:
        container: ContainerProxy instance
        new_max_throughput: New maximum throughput in RU/s
    """
    print("\nUpdate Autoscale Maximum Throughput (Async)")
    print("=" * 70)
    
    try:
        # Update autoscale max throughput
        new_throughput = ThroughputProperties(
            auto_scale_max_throughput=new_max_throughput,
            auto_scale_increment_percent=0
        )
        
        updated_offer = await container.replace_throughput(new_throughput)
        
        autopilot_settings = updated_offer.properties.get('content', {}).get('offerAutopilotSettings')
        if autopilot_settings:
            max_throughput = autopilot_settings.get('maxThroughput')
            print(f"Container '{container.id}' autoscale updated:")
            print(f"  - New maximum throughput: {max_throughput} RU/s")
            print(f"  - New minimum throughput: {max_throughput // 10} RU/s")
            print(f"  - Autoscale will now scale within this new range")
        else:
            print(f"Warning: Updated offer does not contain autoscale settings")
        
    except exceptions.CosmosHttpResponseError as e:
        print(f"Error updating autoscale throughput: {e.message}")

async def run_sample():
    """
    Run the async autoscale throughput management sample.
    """
    print('=' * 70)
    print('Azure Cosmos DB - Async Autoscale Throughput Management Sample')
    print('=' * 70)
    
    # Initialize async client
    async with CosmosClient(HOST, {'masterKey': MASTER_KEY}) as client:
        try:
            # 1. Create database with autoscale
            database = await create_database_with_autoscale(client, DATABASE_ID + '_autoscale_async')
            
            # 2. Create container with autoscale
            container = await create_container_with_autoscale(database, CONTAINER_ID + '_autoscale_async')
            
            # 3. Read autoscale settings
            await read_autoscale_throughput(database, container)
            
            # 4. Update autoscale max throughput
            await update_autoscale_max_throughput(container, 6000)
            
            # 5. Read updated settings
            await read_autoscale_throughput(database, container)

            # Cleanup
            print("\n" + "=" * 70)
            print("Cleaning up resources...")
            print("=" * 70)
            
            await database.delete_container(container.id)
            print(f"Deleted container: {container.id}")
            
            await client.delete_database(database.id)
            print(f"Deleted database: {database.id}")
            
        except exceptions.CosmosHttpResponseError as e:
            print(f"\nError: {e.message}")
    
    print('\n' + '=' * 70)
    print('Sample completed!')
    print('=' * 70)


if __name__ == '__main__':
    asyncio.run(run_sample())



# The MIT License (MIT)
# Copyright (c) 2014 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Diagnostic tools for Azure Cosmos database service operations.
IMPORTANT: This file has been marked for deprecation and will be removed in the future. For diagnostics logging in our
SDK, please use our CosmosHttpLoggingPolicy outlined in our README.
"""
import warnings

from azure.core.utils import CaseInsensitiveDict


class _RecordDiagnostics(object):
    """This file is currently deprecated and will be removed in the future. Please use our CosmosHttpLoggingPolicy
    for logging SDK diagnostics moving forward. More information on this can be found in our README.

    Record Response headers from Cosmos read operations.

    The full response headers are stored in the ``headers`` property.

    Examples:

        >>> rh = RecordDiagnostics()

        >>> col = b.create_container(
        ...     id="some_container",
        ...     partition_key=PartitionKey(path='/id', kind='Hash'),
        ...     response_hook=rh)

        >>> rh.headers['x-ms-activity-id']
        '6243eeed-f06a-413d-b913-dcf8122d0642'

    """

    _common = {
        "x-ms-activity-id",
        "x-ms-session-token",
        "x-ms-item-count",
        "x-ms-request-quota",
        "x-ms-resource-usage",
        "x-ms-retry-after-ms",
    }

    def __init__(self):
        self._headers = CaseInsensitiveDict()
        self._body = None
        self._request_charge = 0

    @property
    def headers(self):
        return CaseInsensitiveDict(self._headers)

    @property
    def body(self):
        return self._body

    @property
    def request_charge(self):
        return self._request_charge

    def clear(self):
        self._request_charge = 0

    def __call__(self, headers, body):
        self._headers = headers
        self._body = body

        self._request_charge += float(headers.get("x-ms-request-charge", 0))

    def __getattr__(self, name):
        key = "x-ms-" + name.replace("_", "-")
        if key in self._common:
            return self._headers[key]
        raise AttributeError(name)


def __getattr__(name):
    if name == 'RecordDiagnostics':
        warnings.warn(
            "RecordDiagnostics is deprecated and should not be used. " +
            "For logging diagnostics information for the SDK, please use our CosmosHttpLoggingPolicy. " +
            "For more information on this, please see our README.",
            DeprecationWarning
        )
        return _RecordDiagnostics

    raise AttributeError(f"module 'azure.cosmos.diagnostics' has no attribute {name}")
